import os
from airunner.enums import SignalCode
from airunner.utils.models.scan_path_for_items import scan_path_for_embeddings


class EmbeddingMixin:
    @property
    def __embeddings(self):
        return self.get_embeddings_by_version(self.generator_settings.version)

    def get_embeddings(self, message: dict = None):
        name_filter = message.get("name_filter") if message is not None else ""
        embeddings = []

        for embedding in self.__embeddings:
            if name_filter == "":
                embeddings.append(embedding)
                continue
            if name_filter in embedding["name"]:
                embeddings.append(embedding)
        self.emit_signal(
            SignalCode.EMBEDDING_GET_ALL_RESULTS_SIGNAL,
            {
                "embeddings": embeddings
            }
        )
        return embeddings

    def delete_missing_embeddings(self):
        embeddings = self.get_embeddings()
        for embedding in embeddings:
            if not os.path.exists(embedding["path"]):
                self.delete_embedding(embedding)
    
    def delete_embedding(self, embedding):
        for index, _embedding in enumerate(self.embeddings):
            if _embedding.name == embedding.name and _embedding.path == embedding.path:
                self.delete_embedding(embedding)
                return

    def scan_for_embeddings(self):
        embeddings = scan_path_for_embeddings(self.path_settings.base_path)
        self.update_embeddings(embeddings)
        self.emit_signal(
            SignalCode.EMBEDDING_GET_ALL_RESULTS_SIGNAL,
            {
                "embeddings": self.get_embeddings_by_version(self.generator_settings.version)
            }
        )



import datetime
from typing import List

from sqlalchemy.orm import joinedload

from airunner.aihandler.models.database_handler import DatabaseHandler
from airunner.aihandler.models.settings_models import Chatbot, AIModels, Schedulers, Lora, PathSettings, SavedPrompt, \
    Embedding, PromptTemplate, ControlnetModel, FontSetting, PipelineModel, ShortcutKeys, \
    GeneratorSettings, WindowSettings, ApplicationSettings, ActiveGridSettings, ControlnetSettings, \
    ImageToImageSettings, OutpaintSettings, DrawingPadSettings, MetadataSettings, \
    LLMGeneratorSettings, TTSSettings, SpeechT5Settings, EspeakSettings, STTSettings, BrushSettings, GridSettings, \
    MemorySettings, Message, Conversation, Summary


class SettingsDBHandler(DatabaseHandler):
    #######################################
    ### SCHEDULERS ###
    #######################################
    def load_schedulers(self) -> List[Schedulers]:
        session = self.get_db_session()
        try:
            return session.query(Schedulers).all()
        finally:
            session.close()

    #######################################
    ### SETTINGS ###
    #######################################
    def load_settings_from_db(self, model_class_):
        session = self.get_db_session()
        try:
            settings = session.query(model_class_).first()
            if settings is None:
                settings = self.create_new_settings(model_class_)
        finally:
            session.close()
        return settings

    def update_setting(self, model_class_, name, value):
        session = self.get_db_session()
        try:
            setting = session.query(model_class_).order_by(model_class_.id.desc()).first()
            if setting:
                setattr(setting, name, value)
                session.commit()
        finally:
            session.close()

    def save_generator_settings(self, generator_settings: GeneratorSettings):
        session = self.get_db_session()
        try:
            query = session.query(GeneratorSettings).filter_by(
                id=generator_settings.id
            ).first()
            if query:
                for key in generator_settings.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(generator_settings, key))
            else:
                session.add(generator_settings)
            session.commit()
        finally:
            session.close()

    def reset_settings(self):
        session = self.get_db_session()
        try:
            # Delete all entries from the model class
            session.query(ApplicationSettings).delete()
            session.query(ActiveGridSettings).delete()
            session.query(ControlnetSettings).delete()
            session.query(ImageToImageSettings).delete()
            session.query(OutpaintSettings).delete()
            session.query(DrawingPadSettings).delete()
            session.query(MetadataSettings).delete()
            session.query(GeneratorSettings).delete()
            session.query(LLMGeneratorSettings).delete()
            session.query(TTSSettings).delete()
            session.query(SpeechT5Settings).delete()
            session.query(EspeakSettings).delete()
            session.query(STTSettings).delete()
            session.query(BrushSettings).delete()
            session.query(GridSettings).delete()
            session.query(PathSettings).delete()
            session.query(MemorySettings).delete()
            # Commit the changes
            session.commit()
        finally:
            session

    def create_new_settings(self, model_class_):
        session = self.get_db_session()
        try:
            new_settings = model_class_()
            session.add(new_settings)
            session.commit()
            session.refresh(new_settings)
        finally:
            session.close()
        return new_settings

    #######################################
    ### SAVED PROMPTS ###
    #######################################
    def get_saved_prompt_by_id(self, prompt_id) -> SavedPrompt:
        session = self.get_db_session()
        try:
            return session.query(SavedPrompt).filter_by(id=prompt_id).first()
        finally:
            session.close()

    def update_saved_prompt(self, saved_prompt: SavedPrompt):
        session = self.get_db_session()
        try:
            query = session.query(SavedPrompt).filter_by(
                id=saved_prompt.id
            ).first()
            if query:
                for key in saved_prompt.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(saved_prompt, key))
            else:
                session.add(saved_prompt)
            session.commit()
        finally:
            session.close()

    def create_saved_prompt(self, data: dict):
        session = self.get_db_session()
        try:
            new_saved_prompt = SavedPrompt(**data)
            session.add(new_saved_prompt)
            session.commit()
        finally:
            session.close()

    def load_saved_prompts(self) -> List[SavedPrompt]:
        session = self.get_db_session()
        try:
            return session.query(SavedPrompt).all()
        finally:
            session.close()

    def load_font_settings(self) -> List[FontSetting]:
        session = self.get_db_session()
        try:
            return session.query(FontSetting).all()
        finally:
            session.close()

    def get_font_setting_by_name(self, name) -> FontSetting:
        session = self.get_db_session()
        try:
            return session.query(FontSetting).filter_by(name=name).first()
        finally:
            session.close()

    def update_font_setting(self, font_setting: FontSetting):
        session = self.get_db_session()
        try:
            query = session.query(FontSetting).filter_by(
                name=font_setting.name
            ).first()
            if query:
                for key in font_setting.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(font_setting, key))
            else:
                session.add(font_setting)
            session.commit()
        finally:
            session.close()

    #######################################
    ### AI MODELS ###
    #######################################
    def load_ai_models(self) -> List[AIModels]:
        session = self.get_db_session()
        try:
            return session.query(AIModels).all()
        finally:
            session.close()

    def update_ai_models(self, models: List[AIModels]):
        for model in models:
            self.update_ai_model(model)

    def update_ai_model(self, model: AIModels):
        session = self.get_db_session()
        try:
            query = session.query(AIModels).filter_by(
                name=model.name,
                path=model.path,
                branch=model.branch,
                version=model.version,
                category=model.category,
                pipeline_action=model.pipeline_action,
                enabled=model.enabled,
                model_type=model.model_type,
                is_default=model.is_default
            ).first()
            if query:
                for key in model.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(model, key))
            else:
                session.add(model)
            session.commit()
        finally:
            session.close()

    #######################################
    ### CHATBOTS ###
    #######################################
    def load_chatbots(self) -> List[Chatbot]:
        session = self.get_db_session()
        try:
            settings = session.query(Chatbot).all()
            return settings
        finally:
            session.close()

    def delete_chatbot_by_name(self, chatbot_name):
        session = self.get_db_session()
        try:
            session.query(Chatbot).filter_by(name=chatbot_name).delete()
            session.commit()
        finally:
            session.close()

    def create_chatbot(self, chatbot_name):
        session = self.get_db_session()
        try:
            new_chatbot = Chatbot(name=chatbot_name)
            session.add(new_chatbot)
            session.commit()
        finally:
            session.close()

    def reset_path_settings(self):
        session = self.get_db_session()
        try:
            # Delete all entries from PathSettings
            session.query(PathSettings).delete()

            # Create a new PathSettings instance with default values
            self.set_default_values(PathSettings)
            # Commit the changes
            session.commit()
        finally:
            session.close()

    def set_default_values(self, model_name_):
        session = self.get_db_session()
        try:
            default_values = {}
            for column in model_name_.__table__.columns:
                if column.default is not None:
                    default_values[column.name] = column.default.arg
            session.execute(
                model_name_.__table__.insert(),
                [default_values]
            )
            session.commit()
        finally:
            session.close()

    #######################################
    ### LORA ###
    #######################################
    def load_lora(self) -> List[Lora]:
        session = self.get_db_session()
        try:
            return session.query(Lora).all()
        finally:
            session.close()

    def get_lora_by_name(self, name):
        session = self.get_db_session()
        try:
            return session.query(Lora).filter_by(name=name).first()
        finally:
            session.close()


    def add_lora(self, lora: Lora):
        session = self.get_db_session()
        try:
            session.add(lora)
            session.commit()
        finally:
            session.close()

    def delete_lora(self, lora: Lora):
        session = self.get_db_session()
        try:
            session.query(Lora).filter_by(name=lora.name).delete()
            session.commit()
        finally:
            session.close()

    def update_lora(self, lora: Lora):
        session = self.get_db_session()
        try:
            query = session.query(Lora).filter_by(name=lora.name).first()
            if query:
                for key in lora.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(lora, key))
            else:
                session.add(lora)
            session.commit()
        finally:
            session.close()

    def update_loras(self, loras: List[Lora]):
        session = self.get_db_session()
        try:
            for lora in loras:
                query = session.query(Lora).filter_by(name=lora.name).first()
                if query:
                    for key in lora.__dict__.keys():
                        if key != "_sa_instance_state":
                            setattr(query, key, getattr(lora, key))
                else:
                    session.add(lora)
            session.commit()
        finally:
            session.close()

    def create_lora(self, lora: Lora):
        session = self.get_db_session()
        try:
            session.add(lora)
            session.commit()
        finally:
            session.close()

    def delete_lora_by_name(self, lora_name, version):
        session = self.get_db_session()
        try:
            session.query(Lora).filter_by(name=lora_name, version=version).delete()
            session.commit()
        finally:
            session.close()

    #######################################
    ### EMBEDDINGS ###
    #######################################
    def delete_embedding(self, embedding: Embedding):
        session = self.get_db_session()
        try:
            session.query(Embedding).filter_by(
                name=embedding.name,
                path=embedding.path,
                branch=embedding.branch,
                version=embedding.version,
                category=embedding.category,
                pipeline_action=embedding.pipeline_action,
                enabled=embedding.enabled,
                model_type=embedding.model_type,
                is_default=embedding.is_default
            ).delete()
            session.commit()
        finally:
            session.close()

    def load_embeddings(self) -> List[Embedding]:
        session = self.get_db_session()
        try:
            return session.query(Embedding).all()
        finally:
            session.close()

    def update_embeddings(self, embeddings: List[Embedding]):
        session = self.get_db_session()
        try:
            for embedding in embeddings:
                query = session.query(Embedding).filter_by(
                    name=embedding.name,
                    path=embedding.path,
                    branch=embedding.branch,
                    version=embedding.version,
                    category=embedding.category,
                    pipeline_action=embedding.pipeline_action,
                    enabled=embedding.enabled,
                    model_type=embedding.model_type,
                    is_default=embedding.is_default
                ).first()
                if query:
                    for key in embedding.__dict__.keys():
                        if key != "_sa_instance_state":
                            setattr(query, key, getattr(embedding, key))
                else:
                    session.add(embedding)
            session.commit()
        finally:
            session.close()

    def get_embedding_by_name(self, name):
        session = self.get_db_session()
        try:
            return session.query(Embedding).filter_by(name=name).first()
        finally:
            session.close()

    def add_embedding(self, embedding: Embedding):
        session = self.get_db_session()
        try:
            session.add(embedding)
            session.commit()
        finally:
            session.close()

    #######################################
    ### PROMPT TEMPLATES ###
    #######################################
    def load_prompt_templates(self) -> List[PromptTemplate]:
        session = self.get_db_session()
        try:
            return session.query(PromptTemplate).all()
        finally:
            session.close()

    def get_prompt_template_by_name(self, name) -> PromptTemplate:
        session = self.get_db_session()
        try:
            return session.query(PromptTemplate).filter_by(template_name=name).first()
        finally:
            session.close()


    #######################################
    ### CONTROLNET MODELS ###
    #######################################
    def load_controlnet_models(self) -> List[ControlnetModel]:
        session = self.get_db_session()
        try:
            return session.query(ControlnetModel).all()
        finally:
            session.close()

    def controlnet_model_by_name(self, name) -> ControlnetModel:
        session = self.get_db_session()
        try:
            return session.query(ControlnetModel).filter_by(name=name).first()
        finally:
            session.close()

    def load_pipelines(self) -> List[PipelineModel]:
        session = self.get_db_session()
        try:
            return session.query(PipelineModel).all()
        finally:
            session.close()


    def load_shortcut_keys(self) -> List[ShortcutKeys]:
        session = self.get_db_session()
        try:
            return session.query(ShortcutKeys).all()
        finally:
            session.close()


    def load_window_settings(self) -> WindowSettings:
        session = self.get_db_session()
        try:
            return session.query(WindowSettings).first()
        finally:
            session.close()


    def save_window_settings(self, window_settings: WindowSettings):
        session = self.get_db_session()
        try:
            query = session.query(WindowSettings).first()
            if query:
                for key in window_settings.__dict__.keys():
                    if key != "_sa_instance_state":
                        setattr(query, key, getattr(window_settings, key))
            else:
                session.add(window_settings)
            session.commit()
        finally:
            session.close()

    def save_object(self, database_object):
        session = self.get_db_session()
        session.add(database_object)
        session.commit()
        session.close()


    def load_history_from_db(self, conversation_id):
        with self.get_db_session() as session:
            messages = session.query(Message).filter_by(
                conversation_id=conversation_id
            ).order_by(Message.timestamp).all()
            results = [
                {
                    "role": message.role,
                    "content": message.content,
                    "name": message.name,
                    "is_bot": message.is_bot,
                    "timestamp": message.timestamp,
                    "conversation_id": message.conversation_id
                } for message in messages
            ]
            return results

    def add_message_to_history(self, content, role, name, is_bot, conversation_id):
        timestamp = datetime.datetime.now()  # Ensure timestamp is a datetime object
        with self.get_db_session() as session:
            llm_generator_settings = session.query(LLMGeneratorSettings).first()
            message = Message(
                role=role,
                content=content,
                name=name,
                is_bot=is_bot,
                timestamp=timestamp,
                conversation_id=conversation_id,
                chatbot_id=llm_generator_settings.current_chatbot
            )
            session.add(message)
            session.commit()

    def get_chatbot_by_id(self, chatbot_id) -> Chatbot:
        session = self.get_db_session()
        try:
            chatbot = session.query(Chatbot).filter_by(id=chatbot_id).options(joinedload(Chatbot.target_files)).first()
            if chatbot is None:
                chatbot = session.query(Chatbot).options(joinedload(Chatbot.target_files)).first()
        finally:
            session.close()
        return chatbot

    def create_conversation(self):
        with self.get_db_session() as session:
            conversation = Conversation(
                timestamp=datetime.datetime.now(datetime.timezone.utc),
                title=""
            )
            session.add(conversation)
            session.commit()
            return conversation.id

    def update_conversation_title(self, conversation_id, title):
        with self.get_db_session() as session:
            conversation = session.query(Conversation).filter_by(id=conversation_id).first()
            if conversation:
                conversation.title = title
                session.commit()

    def add_summary(self, content, conversation_id):
        timestamp = datetime.datetime.now()  # Ensure timestamp is a datetime object
        with self.get_db_session() as session:
            summary = Summary(
                content=content,
                timestamp=timestamp,
                conversation_id=conversation_id
            )
            session.add(summary)
            session.commit()

    def create_conversation_with_messages(self, messages):
        conversation_id = self.create_conversation()
        for message in messages:
            self.add_message_to_history(
                content=message["content"],
                role=message["role"],
                name=message["name"],
                is_bot=message["is_bot"],
                conversation_id=conversation_id
            )
        return conversation_id

    def get_all_conversations(self):
        session = self.Session()
        conversations = session.query(Conversation).all()
        session.close()
        return conversations

    def delete_conversation(self, conversation_id):
        session = self.Session()
        try:
            session.query(Message).filter_by(conversation_id=conversation_id).delete()
            session.query(Summary).filter_by(conversation_id=conversation_id).delete()
            session.query(Conversation).filter_by(id=conversation_id).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            print(f"Error deleting conversation: {e}")
        finally:
            session.close()

    def get_most_recent_conversation_id(self):
        session = self.Session()
        conversation = session.query(Conversation).order_by(Conversation.timestamp.desc()).first()
        session.close()
        return conversation.id if conversation else None



pipeline_bootstrap_data = [
    {
        "pipeline_action": "safety_checker",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "diffusers.pipelines.stable_diffusion.StableDiffusionSafetyChecker",
        "default": True
    },
    {
        "pipeline_action": "controlnet",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "diffusers.ControlNetModel",
        "default": True
    },
    {
        "pipeline_action": "txt2img",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "diffusers.AutoPipelineForText2Image",
        "default": True
    },
    {
        "pipeline_action": "txt2img",
        "version": "SD 1.5",
        "category": "controlnet",
        "classname": "diffusers.StableDiffusionControlNetPipeline",
        "default": False
    },
    {
        "pipeline_action": "txt2img",
        "version": "SDXL 1.0",
        "category": "stablediffusion",
        "classname": "diffusers.AutoPipelineForText2Image",
        "default": False
    },
    {
        "pipeline_action": "txt2img",
        "version": "SDXL Turbo",
        "category": "stablediffusion",
        "classname": "diffusers.AutoPipelineForText2Image",
        "default": False
    },
    {
        "pipeline_action": "img2img",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "diffusers.AutoPipelineForImage2Image",
        "default": False
    },
    {
        "pipeline_action": "img2img",
        "version": "SD 1.5",
        "category": "controlnet",
        "classname": "diffusers.StableDiffusionControlNetImg2ImgPipeline",
        "default": False
    },
    {
        "pipeline_action": "img2img",
        "version": "SDXL 1.0",
        "category": "stablediffusion",
        "classname": "diffusers.AutoPipelineForImage2Image",
        "default": False
    },
    {
        "pipeline_action": "outpaint",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "diffusers.StableDiffusionInpaintPipeline",
        "default": False
    },
    {
        "pipeline_action": "outpaint",
        "version": "SD 1.5",
        "category": "conrolnet",
        "classname": "diffusers.StableDiffusionControlNetInpaintPipeline",
        "default": False
    },
    {
        "pipeline_action": "inpaint_vae",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "diffusers.AsymmetricAutoencoderKL",
        "default": False
    },
    {
        "pipeline_action": "text_encoder",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "transformers.CLIPTextModel",
        "default": False
    },
    {
        "pipeline_action": "text_encoder",
        "version": "SDXL 1.0",
        "category": "stablediffusion",
        "classname": "transformers.CLIPTextModel",
        "default": False
    },
    {
        "pipeline_action": "feature_extractor",
        "version": "SD 1.5",
        "category": "stablediffusion",
        "classname": "transformers.AutoFeatureExtractor",
        "default": False
    },
    {
        "pipeline_action": "seq2seq",
        "version": "1",
        "category": "llm",
        "classname": "transformers.AutoModelForSeq2SeqLM",
        "default": False
    },
    {
        "pipeline_action": "causallm",
        "version": "1",
        "category": "llm",
        "classname": "transformers.AutoModelForCausalLM",
        "default": False
    },
    {
        "pipeline_action": "causallm",
        "version": "0.1",
        "category": "llm",
        "classname": "transformers.AutoModelForCausalLM",
        "default": False
    },
    {
        "pipeline_action": "causallm",
        "version": "2",
        "category": "llm",
        "classname": "transformers.AutoModelForCausalLM",
        "default": False
    }
]



import logging
from typing import List

from sqlalchemy.orm import joinedload

from airunner.aihandler.models.settings_db_handler import SettingsDBHandler
from airunner.aihandler.models.settings_models import ApplicationSettings, LLMGeneratorSettings, GeneratorSettings, \
    ControlnetSettings, BrushSettings, DrawingPadSettings, GridSettings, ActiveGridSettings, \
    ImageToImageSettings, OutpaintSettings, PathSettings, MemorySettings, Chatbot, \
    AIModels, Schedulers, Lora, ShortcutKeys, SavedPrompt, SpeechT5Settings, TTSSettings, EspeakSettings, \
    MetadataSettings, Embedding, STTSettings, PromptTemplate, ControlnetModel, FontSetting, PipelineModel, TargetFiles, \
    ImageFilterValue
from airunner.enums import SignalCode
from airunner.utils.convert_base64_to_image import convert_base64_to_image


class SettingsMixin:
    def __init__(self):
        logging.debug("Initializing SettingsMixin instance")
        self.db_handler = SettingsDBHandler()

    @property
    def stt_settings(self) -> STTSettings:
        return self.db_handler.load_settings_from_db(STTSettings)

    @property
    def application_settings(self) -> ApplicationSettings:
        return self.db_handler.load_settings_from_db(ApplicationSettings)

    @property
    def llm_generator_settings(self) -> LLMGeneratorSettings:
        return self.db_handler.load_settings_from_db(LLMGeneratorSettings)

    @property
    def generator_settings(self) -> GeneratorSettings:
        return self.db_handler.load_settings_from_db(GeneratorSettings)

    @property
    def controlnet_settings(self) -> ControlnetSettings:
        return self.db_handler.load_settings_from_db(ControlnetSettings)

    @property
    def image_to_image_settings(self) -> ImageToImageSettings:
        return self.db_handler.load_settings_from_db(ImageToImageSettings)

    @property
    def outpaint_settings(self) -> OutpaintSettings:
        return self.db_handler.load_settings_from_db(OutpaintSettings)

    @property
    def drawing_pad_settings(self) -> DrawingPadSettings:
        return self.db_handler.load_settings_from_db(DrawingPadSettings)

    @property
    def brush_settings(self) -> BrushSettings:
        return self.db_handler.load_settings_from_db(BrushSettings)

    @property
    def grid_settings(self) -> GridSettings:
        return self.db_handler.load_settings_from_db(GridSettings)

    @property
    def active_grid_settings(self) -> ActiveGridSettings:
        return self.db_handler.load_settings_from_db(ActiveGridSettings)

    @property
    def path_settings(self) -> PathSettings:
        return self.db_handler.load_settings_from_db(PathSettings)

    @property
    def memory_settings(self) -> MemorySettings:
        return self.db_handler.load_settings_from_db(MemorySettings)

    @property
    def chatbots(self) -> List[Chatbot]:
        return self.db_handler.load_chatbots()

    @property
    def ai_models(self) -> List[AIModels]:
        return self.db_handler.load_ai_models()

    @property
    def schedulers(self) -> List[Schedulers]:
        return self.db_handler.load_schedulers()

    @property
    def lora(self) -> List[Lora]:
        return self.db_handler.load_lora()

    @property
    def shortcut_keys(self) -> List[ShortcutKeys]:
        return self.db_handler.load_shortcut_keys()

    @property
    def speech_t5_settings(self) -> SpeechT5Settings:
        return self.db_handler.load_settings_from_db(SpeechT5Settings)

    @property
    def tts_settings(self) -> TTSSettings:
        return self.db_handler.load_settings_from_db(TTSSettings)

    @property
    def espeak_settings(self) -> EspeakSettings:
        return self.db_handler.load_settings_from_db(EspeakSettings)

    @property
    def metadata_settings(self) -> MetadataSettings:
        return self.db_handler.load_settings_from_db(MetadataSettings)

    @property
    def embeddings(self) -> List[Embedding]:
        return self.db_handler.load_embeddings()

    @property
    def prompt_templates(self) -> List[PromptTemplate]:
        return self.db_handler.load_prompt_templates()

    @property
    def controlnet_models(self):
        return self.db_handler.load_controlnet_models()

    @property
    def saved_prompts(self) -> List[SavedPrompt]:
        return self.db_handler.load_saved_prompts()

    @property
    def font_settings(self) -> List[FontSetting]:
        return self.db_handler.load_font_settings()

    @property
    def pipelines(self) -> List[PipelineModel]:
        return self.db_handler.load_pipelines()

    @property
    def drawing_pad_image(self):
        base_64_image = self.drawing_pad_settings.image
        image = convert_base64_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def drawing_pad_mask(self):
        base_64_image = self.drawing_pad_settings.mask
        image = convert_base64_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def img2img_image(self):
        base_64_image = self.image_to_image_settings.image
        image = convert_base64_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def controlnet_image(self):
        base_64_image = self.controlnet_settings.image
        image = convert_base64_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def controlnet_generated_image(self):
        base_64_image = self.controlnet_settings.imported_image_base64
        image = convert_base64_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def outpaint_mask(self):
        base_64_image = self.drawing_pad_settings.mask
        image = convert_base64_to_image(base_64_image)
        if image is not None:
            image = image.convert("RGB")
        return image

    @property
    def image_filter_values(self):
        session = self.db_handler.get_db_session()
        try:
            return session.query(ImageFilterValue).all()
        finally:
            session.close()

    #######################################
    ### LORA ###
    #######################################

    def get_lora_by_version(self, version):
        session = self.db_handler.get_db_session()
        try:
            return session.query(Lora).filter_by(version=version).all()
        finally:
            session.close()

    def delete_lora_by_name(self, name, version):
        self.db_handler.delete_lora_by_name(name, version)

    def create_lora(self, lora: Lora):
        self.db_handler.create_lora(lora)

    def update_lora(self, lora: Lora):
        self.db_handler.update_lora(lora)
        self.__settings_updated()

    def update_loras(self, loras: List[Lora]):
        self.db_handler.update_loras(loras)
        self.__settings_updated()

    #######################################
    ### EMBEDDINGS ###
    #######################################
    def get_embeddings_by_version(self, version):
        return [embedding for embedding in self.embeddings if embedding.version == version]

    def delete_embedding(self, embedding: Embedding):
        self.db_handler.delete_embedding(embedding)

    def update_embeddings(self, embeddings: List[Embedding]):
        self.db_handler.update_embeddings(embeddings)
        self.__settings_updated()

    #######################################
    ### CHATBOT ###
    #######################################
    @property
    def chatbot(self) -> Chatbot:
        return self.get_chatbot_by_id(
            self.llm_generator_settings.current_chatbot
        )

    def get_chatbot_by_id(self, chatbot_id) -> Chatbot:
        chatbot = None
        session = self.db_handler.get_db_session()
        try:
            chatbot = session.query(Chatbot).filter_by(id=chatbot_id).options(joinedload(Chatbot.target_files)).first()
            if chatbot is None:
                chatbot = session.query(Chatbot).options(joinedload(Chatbot.target_files)).first()
        finally:
            session.close()
        return chatbot

    def delete_chatbot_by_name(self, chatbot_name):
        self.db_handler.delete_chatbot_by_name(chatbot_name)

    def create_chatbot(self, chatbot_name):
        self.db_handler.create_chatbot(chatbot_name)

    def add_chatbot_document_to_chatbot(self, chatbot, file_path):
        session = self.db_handler.get_db_session()
        try:
            document = session.query(TargetFiles).filter_by(chatbot_id=chatbot.id, file_path=file_path).first()
            if document is None:
                document = TargetFiles(file_path=file_path, chatbot_id=chatbot.id)
            session.merge(document)  # Use merge instead of add
            session.commit()
        finally:
            session.close()

    #######################################
    ### SAVED PROMPTS ###
    #######################################
    def create_saved_prompt(self, data: dict):
        self.db_handler.create_saved_prompt(data)

    def update_saved_prompt(self, saved_prompt: SavedPrompt):
        self.db_handler.update_saved_prompt(saved_prompt)
        self.__settings_updated()

    def get_saved_prompt_by_id(self, prompt_id) -> SavedPrompt:
        self.db_handler.get_saved_prompt_by_id(prompt_id)

    #######################################
    ### SETTINGS ###
    #######################################
    def update_settings_by_name(self, setting_name, column_name, val):
        if setting_name == "application_settings":
            self.update_application_settings(column_name, val)
        elif setting_name == "generator_settings":
            self.update_generator_settings(column_name, val)
        elif setting_name == "controlnet_image_settings":
            self.update_controlnet_image_settings(column_name, val)
        elif setting_name == "brush_settings":
            self.update_brush_settings(column_name, val)
        elif setting_name == "controlnet_settings":
            self.update_controlnet_settings(column_name, val)
        elif setting_name == "image_to_image_settings":
            self.update_image_to_image_settings(column_name, val)
        elif setting_name == "outpaint_settings":
            self.update_outpaint_settings(column_name, val)
        elif setting_name == "drawing_pad_settings":
            self.update_drawing_pad_settings(column_name, val)
        elif setting_name == "grid_settings":
            self.update_grid_settings(column_name, val)
        elif setting_name == "active_grid_settings":
            self.update_active_grid_settings(column_name, val)
        elif setting_name == "path_settings":
            self.update_path_settings(column_name, val)
        elif setting_name == "memory_settings":
            self.update_memory_settings(column_name, val)
        elif setting_name == "llm_generator_settings":
            self.update_llm_generator_settings(column_name, val)
        else:
            logging.error(f"Invalid setting name: {setting_name}")

    def update_application_settings(self, column_name, val):
        self.db_handler.update_setting(ApplicationSettings, column_name, val)
        self.__settings_updated()

    #######################################
    ### TTS Settings ###
    #######################################
    def update_espeak_settings(self, column_name, val):
        self.db_handler.update_setting(EspeakSettings, column_name, val)
        self.__settings_updated()

    def update_tts_settings(self, column_name, val):
        self.db_handler.update_setting(TTSSettings, column_name, val)
        self.__settings_updated()

    def update_speech_t5_settings(self, column_name, val):
        self.db_handler.update_setting(SpeechT5Settings, column_name, val)
        self.__settings_updated()


    #######################################
    ### CONTROLNET ###
    #######################################
    def update_controlnet_settings(self, column_name, val):
        self.db_handler.update_setting(ControlnetSettings, column_name, val)
        self.__settings_updated()

    def update_brush_settings(self, column_name, val):
        self.db_handler.update_setting(BrushSettings, column_name, val)
        self.__settings_updated()

    def update_image_to_image_settings(self, column_name, val):
        self.db_handler.update_setting(ImageToImageSettings, column_name, val)
        self.__settings_updated()

    def update_outpaint_settings(self, column_name, val):
        self.db_handler.update_setting(OutpaintSettings, column_name, val)
        self.__settings_updated()

    def update_drawing_pad_settings(self, column_name, val):
        self.db_handler.update_setting(DrawingPadSettings, column_name, val)
        self.__settings_updated()

    def update_grid_settings(self, column_name, val):
        self.db_handler.update_setting(GridSettings, column_name, val)
        self.__settings_updated()

    def update_active_grid_settings(self, column_name, val):
        self.db_handler.update_setting(ActiveGridSettings, column_name, val)
        self.__settings_updated()

    #######################################
    ### PATH ###
    #######################################
    def update_path_settings(self, column_name, val):
        self.db_handler.update_setting(PathSettings, column_name, val)
        self.__settings_updated()

    def reset_path_settings(self):
        self.db_handler.reset_path_settings()

    def update_memory_settings(self, column_name, val):
        self.db_handler.update_setting(MemorySettings, column_name, val)
        self.__settings_updated()

    def update_metadata_settings(self, column_name, val):
        self.db_handler.update_setting(MetadataSettings, column_name, val)
        self.__settings_updated()

    def update_llm_generator_settings(self, column_name, val):
        self.db_handler.update_setting(LLMGeneratorSettings, column_name, val)
        self.__settings_updated()

    def __settings_updated(self):
        self.emit_signal(SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL)

    def reset_settings(self):
        self.db_handler.reset_settings()

    #######################################
    ### AI MODELS ###
    #######################################
    def update_ai_models(self, models: List[AIModels]):
        self.db_handler.update_ai_models(models)
        self.__settings_updated()

    def update_ai_model(self, model: AIModels):
        self.db_handler.update_ai_model(model)
        self.__settings_updated()

    #######################################
    ### PROMPT TEMPLATES ###
    #######################################
    def get_prompt_template_by_name(self, name) -> PromptTemplate:
        return self.db_handler.get_prompt_template_by_name(name)

    #######################################
    ### CONTROLNET MODELS ###
    #######################################
    def controlnet_model_by_name(self, name) -> ControlnetModel:
        return self.db_handler.controlnet_model_by_name(name)

    def get_font_setting_by_name(self, name) -> FontSetting:
        return self.db_handler.get_font_setting_by_name(name)

    def update_font_setting(self, font_setting: FontSetting):
        self.db_handler.update_font_setting(font_setting)
        self.__settings_updated()

    def ai_model_get_by_filter(self, filter_dict: dict) -> List[AIModels]:
        results = []
        models = self.db_handler.load_ai_models()
        for item in models:
            match = True
            for k, v in filter_dict.items():
                if isinstance(item, dict):
                    if item.get(k, "") != v:
                        match = False
                        break
                else:
                    if not hasattr(item, k) or getattr(item, k) != v:
                        match = False
                        break
            if match:
                results.append(item)
        return results

    #######################################
    ### GENERATOR SETTINGS ###
    #######################################
    def update_generator_settings(self, column_name, val):
        generator_settings = self.generator_settings
        setattr(generator_settings, column_name, val)
        self.save_generator_settings(generator_settings)

    def save_generator_settings(self, generator_settings):
        self.db_handler.save_generator_settings(generator_settings)
        self.__settings_updated()


    #######################################
    ### WINDOW SETTINGS ###
    #######################################
    @property
    def window_settings(self):
        return self.db_handler.load_window_settings()


    def save_window_settings(self, column_name, val):
        window_settings = self.window_settings
        setattr(window_settings, column_name, val)
        self.db_handler.save_window_settings(window_settings)



import os
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from airunner.aihandler.models.settings_models import Base

class DatabaseHandler:
    def __init__(self, db_path=os.path.expanduser(
        os.path.join(
            "~",
            ".local",
            "share",
            "airunner",
            "data",
            "airunner.db"
        )
    )):
        self.db_path = db_path
        self.engine = create_engine(f'sqlite:///{self.db_path}')
        Base.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)
        self.conversation_id = None

    def get_db_session(self):
        return self.Session()



import logging

from airunner.settings import LOG_LEVEL
import warnings
import time


class Logger:
    """
    Wrapper class for logging
    """

    # disable warnings
    warnings.filterwarnings("ignore")

    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    FATAL = logging.FATAL
    
    def __init__(self, **kwargs):
        prefix = kwargs.pop("prefix", "")
        name = kwargs.pop("name", "AI Runner")
        super().__init__()
        self.logger = logging.getLogger(f"{name}::{prefix}")
        self.set_level(LOG_LEVEL)

    @property
    def date_time(self):
        """
        Get the current date time
        :return: str
        """
        return time.strftime("%Y-%m-%d %H:%M:%S")

    def set_level(self, level):
        """
        Set the logging level
        :param level:
        :return: None
        """
        if level is None:
            level = logging.DEBUG
        self.logger.setLevel(level)

    def debug(self, msg):
        """
        Log info message
        :param msg:
        :return: None
        """
        self.logger.debug(f"{self.date_time} - {msg}")

    def info(self, msg):
        """
        Log info message
        :param msg:
        :return: None
        """
        self.logger.debug(f"{self.date_time} - {msg}")

    def warning(self, msg):
        """
        Log warning message
        :param msg:
        :return: None
        """
        self.logger.warning(f"{self.date_time} - {msg}")

    def error(self, msg):
        """
        Log error message
        :param msg:
        :return: None
        """
        self.logger.error(f"{self.date_time} - {msg}")



from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.tts.templates.speecht5_preferences_ui import Ui_speecht5_preferences


class SpeechT5PreferencesWidget(BaseWidget):
    widget_class_ = Ui_speecht5_preferences

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize_form(self):
        self.ui.rate.init(current_value=self.speech_t5_settings.rate)
        self.ui.volume.init(current_value=self.speech_t5_settings.volume)
        self.ui.pitch.init(current_value=self.speech_t5_settings.pitch)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'model_manager.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QScrollArea, QSizePolicy,
    QTabWidget, QWidget)

from airunner.widgets.embeddings.embeddings_container_widget import EmbeddingsContainerWidget
from airunner.widgets.lora.lora_container_widget import LoraContainerWidget
from airunner.widgets.model_manager.custom_widget import CustomModelWidget
from airunner.widgets.model_manager.default_widget import DefaultModelWidget
from airunner.widgets.model_manager.import_widget import ImportWidget

class Ui_model_manager(object):
    def setupUi(self, model_manager):
        if not model_manager.objectName():
            model_manager.setObjectName(u"model_manager")
        model_manager.resize(472, 380)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(model_manager.sizePolicy().hasHeightForWidth())
        model_manager.setSizePolicy(sizePolicy)
        model_manager.setMinimumSize(QSize(0, 0))
        font = QFont()
        font.setPointSize(9)
        model_manager.setFont(font)
        self.gridLayout = QGridLayout(model_manager)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.tabs = QTabWidget(model_manager)
        self.tabs.setObjectName(u"tabs")
        sizePolicy.setHeightForWidth(self.tabs.sizePolicy().hasHeightForWidth())
        self.tabs.setSizePolicy(sizePolicy)
        self.tabs.setFont(font)
        self.tabs.setStyleSheet(u"")
        self.tabs.setTabsClosable(False)
        self.tabs.setTabBarAutoHide(False)
        self.default_tab = QWidget()
        self.default_tab.setObjectName(u"default_tab")
        self.gridLayout_2 = QGridLayout(self.default_tab)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.scrollArea = QScrollArea(self.default_tab)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 448, 333))
        self.gridLayout_5 = QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_5.setObjectName(u"gridLayout_5")
        self.widget_2 = DefaultModelWidget(self.scrollAreaWidgetContents)
        self.widget_2.setObjectName(u"widget_2")

        self.gridLayout_5.addWidget(self.widget_2, 0, 0, 1, 1)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.gridLayout_2.addWidget(self.scrollArea, 0, 0, 1, 1)

        self.tabs.addTab(self.default_tab, "")
        self.custom_tab = QWidget()
        self.custom_tab.setObjectName(u"custom_tab")
        self.gridLayout_3 = QGridLayout(self.custom_tab)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.scrollArea_2 = QScrollArea(self.custom_tab)
        self.scrollArea_2.setObjectName(u"scrollArea_2")
        self.scrollArea_2.setWidgetResizable(True)
        self.scrollAreaWidgetContents_2 = QWidget()
        self.scrollAreaWidgetContents_2.setObjectName(u"scrollAreaWidgetContents_2")
        self.scrollAreaWidgetContents_2.setGeometry(QRect(0, 0, 448, 333))
        self.gridLayout_6 = QGridLayout(self.scrollAreaWidgetContents_2)
        self.gridLayout_6.setObjectName(u"gridLayout_6")
        self.widget = CustomModelWidget(self.scrollAreaWidgetContents_2)
        self.widget.setObjectName(u"widget")

        self.gridLayout_6.addWidget(self.widget, 0, 0, 1, 1)

        self.scrollArea_2.setWidget(self.scrollAreaWidgetContents_2)

        self.gridLayout_3.addWidget(self.scrollArea_2, 0, 0, 1, 1)

        self.tabs.addTab(self.custom_tab, "")
        self.lora_tab = QWidget()
        self.lora_tab.setObjectName(u"lora_tab")
        self.gridLayout_9 = QGridLayout(self.lora_tab)
        self.gridLayout_9.setObjectName(u"gridLayout_9")
        self.scrollArea_4 = QScrollArea(self.lora_tab)
        self.scrollArea_4.setObjectName(u"scrollArea_4")
        self.scrollArea_4.setWidgetResizable(True)
        self.scrollAreaWidgetContents_4 = QWidget()
        self.scrollAreaWidgetContents_4.setObjectName(u"scrollAreaWidgetContents_4")
        self.scrollAreaWidgetContents_4.setGeometry(QRect(0, 0, 448, 333))
        self.gridLayout_8 = QGridLayout(self.scrollAreaWidgetContents_4)
        self.gridLayout_8.setObjectName(u"gridLayout_8")
        self.lora_widget = LoraContainerWidget(self.scrollAreaWidgetContents_4)
        self.lora_widget.setObjectName(u"lora_widget")

        self.gridLayout_8.addWidget(self.lora_widget, 0, 0, 1, 1)

        self.scrollArea_4.setWidget(self.scrollAreaWidgetContents_4)

        self.gridLayout_9.addWidget(self.scrollArea_4, 0, 0, 1, 1)

        self.tabs.addTab(self.lora_tab, "")
        self.embeddings_tab = QWidget()
        self.embeddings_tab.setObjectName(u"embeddings_tab")
        self.gridLayout_11 = QGridLayout(self.embeddings_tab)
        self.gridLayout_11.setObjectName(u"gridLayout_11")
        self.scrollArea_5 = QScrollArea(self.embeddings_tab)
        self.scrollArea_5.setObjectName(u"scrollArea_5")
        self.scrollArea_5.setWidgetResizable(True)
        self.scrollAreaWidgetContents_5 = QWidget()
        self.scrollAreaWidgetContents_5.setObjectName(u"scrollAreaWidgetContents_5")
        self.scrollAreaWidgetContents_5.setGeometry(QRect(0, 0, 448, 333))
        self.gridLayout_10 = QGridLayout(self.scrollAreaWidgetContents_5)
        self.gridLayout_10.setObjectName(u"gridLayout_10")
        self.widget_4 = EmbeddingsContainerWidget(self.scrollAreaWidgetContents_5)
        self.widget_4.setObjectName(u"widget_4")

        self.gridLayout_10.addWidget(self.widget_4, 0, 0, 1, 1)

        self.scrollArea_5.setWidget(self.scrollAreaWidgetContents_5)

        self.gridLayout_11.addWidget(self.scrollArea_5, 0, 0, 1, 1)

        self.tabs.addTab(self.embeddings_tab, "")
        self.import_tab = QWidget()
        self.import_tab.setObjectName(u"import_tab")
        self.gridLayout_4 = QGridLayout(self.import_tab)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.scrollArea_3 = QScrollArea(self.import_tab)
        self.scrollArea_3.setObjectName(u"scrollArea_3")
        self.scrollArea_3.setWidgetResizable(True)
        self.scrollAreaWidgetContents_3 = QWidget()
        self.scrollAreaWidgetContents_3.setObjectName(u"scrollAreaWidgetContents_3")
        self.scrollAreaWidgetContents_3.setGeometry(QRect(0, 0, 448, 333))
        self.gridLayout_7 = QGridLayout(self.scrollAreaWidgetContents_3)
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.widget_3 = ImportWidget(self.scrollAreaWidgetContents_3)
        self.widget_3.setObjectName(u"widget_3")

        self.gridLayout_7.addWidget(self.widget_3, 0, 0, 1, 1)

        self.scrollArea_3.setWidget(self.scrollAreaWidgetContents_3)

        self.gridLayout_4.addWidget(self.scrollArea_3, 0, 0, 1, 1)

        self.tabs.addTab(self.import_tab, "")

        self.gridLayout.addWidget(self.tabs, 0, 0, 1, 1)


        self.retranslateUi(model_manager)
        self.tabs.currentChanged.connect(model_manager.tab_changed)

        self.tabs.setCurrentIndex(4)


        QMetaObject.connectSlotsByName(model_manager)
    # setupUi

    def retranslateUi(self, model_manager):
        model_manager.setWindowTitle(QCoreApplication.translate("model_manager", u"Form", None))
        self.tabs.setTabText(self.tabs.indexOf(self.default_tab), QCoreApplication.translate("model_manager", u"Default", None))
        self.tabs.setTabText(self.tabs.indexOf(self.custom_tab), QCoreApplication.translate("model_manager", u"Imported", None))
        self.tabs.setTabText(self.tabs.indexOf(self.lora_tab), QCoreApplication.translate("model_manager", u"LoRA", None))
        self.tabs.setTabText(self.tabs.indexOf(self.embeddings_tab), QCoreApplication.translate("model_manager", u"Embeddings", None))
        self.tabs.setTabText(self.tabs.indexOf(self.import_tab), QCoreApplication.translate("model_manager", u"Import", None))
    # retranslateUi




from airunner.data.bootstrap.pipeline_bootstrap_data import pipeline_bootstrap_data


class PipelineMixin:
    def pipeline_get_by_filter(self, filter_dict):
        return [item for item in pipeline_bootstrap_data if all(item.get(k) == v for k, v in filter_dict.items())]

    def get_pipeline_classname(self, pipeline_action, version, category):
        pipelines = self.get_pipelines(pipeline_action, version, category)
        if len(pipelines) > 0:
            return pipelines[0]["classname"]
        else:
            return None
    
    def get_pipelines(self, pipeline_action=None, version=None, category=None):
        pipelines = pipeline_bootstrap_data
        if pipeline_action:
            pipelines = self.pipeline_get_by_filter({"pipeline_action": pipeline_action})
        if version:
            pipelines = self.pipeline_get_by_filter({"version": version})
        if category:
            pipelines = self.pipeline_get_by_filter({"category": category})
        return pipelines

    def available_pipeline_by_section(self, section):
        return [pipeline["name"] for pipeline in pipeline_bootstrap_data if pipeline["section"] == section]

    def available_pipeline_by_action_version_category(self, pipeline_action, version, category):
        return [pipeline["name"] for pipeline in pipeline_bootstrap_data if pipeline["pipeline_action"] == pipeline_action and pipeline["version"] == version and pipeline["category"] == category]

    def pipeline_actions(self):
        return [pipeline["pipeline_action"] for pipeline in pipeline_bootstrap_data]



#!/usr/bin/env python
"""
----------------------------------------------------------------
Import order is crucial for AI Runner to work as expected.
Do not remove the no_internet_socket import.
Do not change the order of the imports.
----------------------------------------------------------------
"""
################################################################
# Importing this module sets the Hugging Face environment
# variables for the application.
################################################################
import facehuggershield
from airunner.settings import NLTK_DOWNLOAD_DIR
facehuggershield.huggingface.activate(
    show_stdout=True,
    darklock_os_whitelisted_directories=[
        "~/.local/share/airunner",
        NLTK_DOWNLOAD_DIR,
        "/tmp"
    ]
)

################################################################
# Set the environment variable for PyTorch to use expandable
################################################################
import os
os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"

################################################################
# Ensure that the base directory exists.
################################################################
base_dir = os.path.join(os.path.expanduser("~"), ".local", "share", "airunner", "data")
os.makedirs(base_dir, exist_ok=True)

################################################################
# Import the main application class for AI Runner.
################################################################
from airunner.app import App

################################################################
# Import Alembic modules to run migrations.
################################################################
from alembic.config import Config
from alembic import command

def setup_database():
    alembic_cfg = Config("alembic.ini")
    command.upgrade(alembic_cfg, "head")


if __name__ == "__main__":
    setup_database()
    App(
        restrict_os_access=None,
        defendatron=facehuggershield.huggingface.defendatron
    )



import json
import re
import time
from PIL import Image

from PySide6.QtCore import Signal, QRect, QThread, QObject, Slot
from PySide6.QtWidgets import QApplication

from airunner.aihandler.models.settings_models import ShortcutKeys
from airunner.enums import SignalCode, GeneratorSection, ImageCategory, ImagePreset, StableDiffusionVersion, \
    ModelStatus, ModelType
from airunner.mediator_mixin import MediatorMixin
from airunner.settings import PHOTO_REALISTIC_NEGATIVE_PROMPT, ILLUSTRATION_NEGATIVE_PROMPT
from airunner.utils.random_seed import random_seed
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.generator_form.templates.generatorform_ui import Ui_generator_form
from airunner.windows.main.settings_mixin import SettingsMixin


class SaveGeneratorSettingsWorker(
    QObject,
    MediatorMixin,
    SettingsMixin
):
    def __init__(self, parent):
        MediatorMixin.__init__(self)
        SettingsMixin.__init__(self)
        super().__init__()
        self.parent = parent
        self.current_prompt_value = None
        self.current_negative_prompt_value = None
        self.current_secondary_prompt_value = None
        self.current_secondary_negative_prompt_value = None
        self.crops_coord_top_left_x = 0
        self.crops_coord_top_left_y = 0


    def run(self):
        do_update_settings = False
        while True:
            value = self.parent.ui.prompt.toPlainText()
            if value != self.current_prompt_value:
                self.current_prompt_value = value
                do_update_settings = True

            value = self.parent.ui.negative_prompt.toPlainText()
            if value != self.current_negative_prompt_value:
                self.current_negative_prompt_value = value
                do_update_settings = True

            value = self.parent.ui.secondary_prompt.toPlainText()
            if value != self.current_secondary_prompt_value:
                self.current_secondary_prompt_value = value
                do_update_settings = True

            value = self.parent.ui.secondary_negative_prompt.toPlainText()
            if value != self.current_secondary_negative_prompt_value:
                self.current_secondary_negative_prompt_value = value
                do_update_settings = True

            x = self.parent.ui.crops_coord_top_left_x.text()
            y = self.parent.ui.crops_coord_top_left_y.text()
            x = int(x) if x != '' else 0
            y = int(y) if y != '' else 0

            if self.crops_coord_top_left_x != x:
                self.crops_coord_top_left_x = x
                do_update_settings = True
            if self.crops_coord_top_left_y != y:
                self.crops_coord_top_left_y = y
                do_update_settings = True

            if do_update_settings:
                do_update_settings = False
                generator_settings = self.generator_settings
                generator_settings.prompt = self.current_prompt_value
                generator_settings.negative_prompt = self.current_negative_prompt_value
                generator_settings.second_prompt = self.current_secondary_prompt_value
                generator_settings.second_negative_prompt = self.current_secondary_negative_prompt_value
                generator_settings.crops_coord_top_left = dict(
                    x=self.crops_coord_top_left_x,
                    y=self.crops_coord_top_left_y
                )
                self.save_generator_settings(generator_settings)

            time.sleep(0.1)


class GeneratorForm(BaseWidget):
    widget_class_ = Ui_generator_form
    changed_signal = Signal(str, object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seed_override = None
        self.parent = None
        self.initialized = False
        self.showing_past_conversations = False
        self.signal_handlers = {
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed_signal,
            SignalCode.SD_GENERATE_IMAGE_SIGNAL: self.on_generate_image_signal,
            SignalCode.APPLICATION_STOP_SD_PROGRESS_BAR_SIGNAL: self.on_stop_image_generator_progress_bar_signal,
            SignalCode.SD_PROGRESS_SIGNAL: self.on_progress_signal,
            SignalCode.GENERATOR_FORM_UPDATE_VALUES_SIGNAL: self.set_form_values,
            SignalCode.LLM_IMAGE_PROMPT_GENERATED_SIGNAL: self.on_llm_image_prompt_generated_signal,
            SignalCode.GENERATE_IMAGE_FROM_IMAGE_SIGNAL: self.handle_generate_image_from_image,
            SignalCode.DO_GENERATE_IMAGE_FROM_IMAGE_SIGNAL: self.do_generate_image_from_image_signal_handler,
            SignalCode.SD_LOAD_PROMPT_SIGNAL: self.on_load_saved_stablediffuion_prompt_signal,
            SignalCode.LOAD_CONVERSATION: self.on_load_conversation,
            SignalCode.BOT_MOOD_UPDATED: self.on_bot_mood_updated,
            SignalCode.KEYBOARD_SHORTCUTS_UPDATED: self.on_keyboard_shortcuts_updated,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
        }
        self.thread = QThread()
        self.worker = SaveGeneratorSettingsWorker(parent=self)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)

    @property
    def is_txt2img(self):
        return self.pipeline_action == GeneratorSection.TXT2IMG.value

    @property
    def is_outpaint(self):
        return self.pipeline_action == GeneratorSection.OUTPAINT.value

    @property
    def pipeline_action(self):
        return self.generator_settings.pipeline_action

    @property
    def generator_name(self):
        return self.application_settings.current_image_generator

    @property
    def seed(self):
        return self.generator_settings.seed

    @seed.setter
    def seed(self, val):
        self.update_generator_settings("seed", val)

    @property
    def active_rect(self):
        rect = QRect(
            self.active_grid_settings.pos_x,
            self.active_grid_settings.pos_y,
            self.application_settings.working_width,
            self.application_settings.working_height
        )
        rect.translate(-self.drawing_pad_settings.x_pos, -self.drawing_pad_settings.y_pos)

        return rect

    def on_keyboard_shortcuts_updated(self):
        self._set_keyboard_shortcuts()

    def on_application_settings_changed_signal(self, _data):
        self.toggle_secondary_prompts()

    def on_bot_mood_updated(self):
        self._set_chatbot_mood()

    def on_generate_image_signal(self, _data):
        self.handle_generate_button_clicked()

    def on_stop_image_generator_progress_bar_signal(self, _data):
        self.stop_progress_bar()

    def on_progress_signal(self, message):
        self.handle_progress_bar(message)

    ##########################################################################
    # LLM Generated Image handlers
    ##########################################################################
    def on_llm_image_prompt_generated_signal(self, data):
        """
        This slot is called after an LLM has generated the prompts for an image.
        It sets the prompts in the generator form UI and continues the image generation process.
        """

        # Send a messagae to the user as chatbot letting them know that the image is generating
        self.emit_signal(
            SignalCode.LLM_TEXT_STREAMED_SIGNAL,
            dict(
                message="Your image is generating...",
                is_first_message=True,
                is_end_of_message=True,
                name=self.chatbot.name
            )
        )

        # Unload the LLM
        if self.application_settings.llm_enabled:
            self.emit_signal(SignalCode.TOGGLE_LLM_SIGNAL, dict(
                callback=self.unload_llm_callback
            ))

        # Set the prompts in the generator form UI
        data = self.extract_json_from_message(data["message"])
        prompt = data.get("composition", None)
        secondary_prompt = data.get("description", None)
        prompt_type = data.get("type", ImageCategory.PHOTO.value)
        if prompt_type == "photo":
            negative_prompt = PHOTO_REALISTIC_NEGATIVE_PROMPT
        else:
            negative_prompt = ILLUSTRATION_NEGATIVE_PROMPT
        self.ui.prompt.setPlainText(prompt)
        self.ui.negative_prompt.setPlainText(negative_prompt)
        self.ui.secondary_prompt.setPlainText(secondary_prompt)
        self.ui.secondary_negative_prompt.setPlainText(negative_prompt)

    def unload_llm_callback(self, data:dict=None):
        """
        Callback function to be called after the LLM has been unloaded.
        """
        if not self.application_settings.sd_enabled:
            # If SD is not enabled, enable it and then emit a signal to generate the image
            # The callback function is handled by the signal handler for the SD_LOAD_SIGNAL.
            # The finalize function is a callback which is called after the image has been generated.
            self.emit_signal(SignalCode.TOGGLE_SD_SIGNAL, dict(
                callback=self.handle_generate_button_clicked,
                finalize=self.finalize_image_generated_by_llm
            ))
        else:
            # If SD is already enabled, emit a signal to generate the image.
            # The finalize function is a callback which is called after the image has been generated.
            self.handle_generate_button_clicked(dict(
                finalize=self.finalize_image_generated_by_llm
            ))

    def finalize_image_generated_by_llm(self, data):
        """
        Callback function to be called after the image has been generated.
        """

        # Create a message to be sent to the user as a chatbot message
        image_generated_message = dict(
            message="Your image has been generated",
            is_first_message=True,
            is_end_of_message=True,
            name=self.chatbot.name
        )

        # If SD is enabled, emit a signal to unload SD.
        if self.application_settings.sd_enabled:
            # If LLM is disabled, emit a signal to load it.
            if not self.application_settings.llm_enabled:
                self.emit_signal(SignalCode.TOGGLE_SD_SIGNAL, dict(
                    callback=lambda d: self.emit_signal(SignalCode.TOGGLE_LLM_SIGNAL, dict(
                        callback=lambda d: self.emit_signal(
                            SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                            image_generated_message
                        )
                    ))
                ))
            else:
                self.emit_signal(SignalCode.TOGGLE_SD_SIGNAL, dict(
                    callback=lambda d: self.emit_signal(
                        SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                        image_generated_message
                    )
                ))
        else:
            # If SD is disabled and LLM is disabled, emit a signal to load LLM
            # with a callback to add the image generated message to the conversation.
            if not self.application_settings.llm_enabled:
                self.emit_signal(SignalCode.TOGGLE_LLM_SIGNAL, dict(
                    callback=lambda d: self.emit_signal(
                        SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                        image_generated_message
                    )
                ))
            else:
                # If SD is disabled and LLM is enabled, emit a signal to add
                # the image generated message to the conversation.
                self.emit_signal(
                    SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                    image_generated_message
                )
    ##########################################################################
    # End LLM Generated Image handlers
    ##########################################################################

    def _set_chatbot_mood(self):
        self.ui.mood_label.setText(self.chatbot.bot_mood)

    def handle_generate_image_from_image(self, image):
        pass

    def on_load_conversation(self, _data):
        self.ui.generator_form_tabs.setCurrentIndex(1)

    def toggle_secondary_prompts(self):
        if self.generator_settings.version != StableDiffusionVersion.SDXL1_0.value:
            self.ui.croops_coord_top_left_groupbox.hide()
        else:
            self.ui.croops_coord_top_left_groupbox.show()

    def on_load_saved_stablediffuion_prompt_signal(self, data: dict):
        saved_prompt = data.get("saved_prompt")
        self.ui.prompt.blockSignals(True)
        self.ui.negative_prompt.blockSignals(True)
        self.ui.secondary_prompt.blockSignals(True)
        self.ui.secondary_negative_prompt.blockSignals(True)
        self.ui.prompt.setPlainText(saved_prompt.prompt)
        self.ui.negative_prompt.setPlainText(saved_prompt.negative_prompt)
        self.ui.secondary_prompt.setPlainText(saved_prompt.secondary_prompt)
        self.ui.secondary_negative_prompt.setPlainText(saved_prompt.secondary_negative_prompt)
        self.ui.prompt.blockSignals(False)
        self.ui.negative_prompt.blockSignals(False)
        self.ui.secondary_prompt.blockSignals(False)
        self.ui.secondary_negative_prompt.blockSignals(False)

    def handle_image_presets_changed(self, val):
        self.update_generator_settings("image_preset", val)

    def do_generate_image_from_image_signal_handler(self, _data):
        self.do_generate()

    def do_generate(self, data=None):
        if data:
            finalize = data.get("finalize", None)
            if finalize:
                data = dict(
                    callback=finalize
                )
            else:
                data = None
        self.emit_signal(SignalCode.DO_GENERATE_SIGNAL, data)

    def activate_ai_mode(self):
        ai_mode = self.application_settings.ai_mode
        self.ui.generator_form_tabs.setCurrentIndex(1 if ai_mode is True else 0)

    def action_clicked_button_save_prompts(self):
        self.emit_signal(SignalCode.SD_SAVE_PROMPT_SIGNAL, {
            "prompt": self.ui.prompt.toPlainText(),
            "negative_prompt": self.ui.negative_prompt.toPlainText(),
            "secondary_prompt": self.ui.secondary_prompt.toPlainText(),
            "secondary_negative_prompt": self.ui.secondary_negative_prompt.toPlainText(),
        })

    def handle_prompt_changed(self):
        pass

    def handle_negative_prompt_changed(self):
        pass

    def handle_second_prompt_changed(self):
        pass

    def handle_second_negative_prompt_changed(self):
        pass

    def handle_generate_button_clicked(self, data=None):
        self.start_progress_bar()
        self.generate(data)

    @Slot()
    def handle_interrupt_button_clicked(self):
        self.emit_signal(SignalCode.INTERRUPT_IMAGE_GENERATION_SIGNAL)

    def generate(self, data=None):
        if self.generator_settings.random_seed:
            self.seed = random_seed()
        self.do_generate(data)
        self.seed_override = None

    def do_generate_image(self):
        time.sleep(0.1)
        self.do_generate()

    def extract_json_from_message(self, message):
        # Regular expression to find the JSON block
        json_pattern = re.compile(r'.*`json\s*({.*?})\s*`.*', re.DOTALL)
        match = json_pattern.search(message)

        if match:
            json_block = match.group(1)
            try:
                # Convert the JSON block to a dictionary
                json_dict = json.loads(json_block)
                return json_dict
            except json.JSONDecodeError as e:
                print(f"Error decoding JSON: {e}")
                return {}
        else:
            print("No JSON block found in the message.")
            return {}

    def get_memory_options(self):
        return {
            "use_last_channels": self.memory_settings.use_last_channels,
            "use_enable_sequential_cpu_offload": self.memory_settings.use_enable_sequential_cpu_offload,
            "enable_model_cpu_offload": self.memory_settings.enable_model_cpu_offload,
            "use_attention_slicing": self.memory_settings.use_attention_slicing,
            "use_tf32": self.memory_settings.use_tf32,
            "use_cudnn_benchmark": self.memory_settings.use_cudnn_benchmark,
            "use_enable_vae_slicing": self.memory_settings.use_enable_vae_slicing,
            "use_accelerated_transformers": self.memory_settings.use_accelerated_transformers,
            "use_torch_compile": self.memory_settings.use_torch_compile,
            "use_tiled_vae": self.memory_settings.use_tiled_vae,
            "use_tome_sd": self.memory_settings.use_tome_sd,
            "tome_sd_ratio": self.memory_settings.tome_sd_ratio,
        }

    def handle_quality_effects_changed(self, val):
        self.update_generator_settings("quality_effects", val)

    def handle_progress_bar(self, message):
        step = message.get("step")
        total = message.get("total")
        if step == total:
            self.stop_progress_bar()
            return

        if step == 0 and total == 0:
            current = 0
        else:
            try:
                current = (step / total)
            except ZeroDivisionError:
                current = 0
        self.set_progress_bar_value(int(current * 100))
    
    def set_progress_bar_value(self, value):
        progressbar = self.ui.progress_bar
        if not progressbar:
            return
        if progressbar.maximum() == 0:
            progressbar.setRange(0, 100)
        progressbar.setValue(value)
        QApplication.processEvents()

    def start_progress_bar(self):
        self.ui.progress_bar.setFormat("Generating %p%")
        self.ui.progress_bar.setRange(0, 0)
        self.ui.progress_bar.show()

    def on_model_status_changed_signal(self, data):
        if data["model"] is ModelType.SD:
            if data["status"] is not ModelStatus.LOADING:
                self.stop_progress_bar(do_clear=True)
                self.ui.generate_button.setEnabled(True)
                self.ui.interrupt_button.setEnabled(True)
            else:
                self.start_progress_bar()
                self.ui.generate_button.setEnabled(False)
                self.ui.interrupt_button.setEnabled(False)

    def showEvent(self, event):
        super().showEvent(event)
        self.activate_ai_mode()
        self.set_form_values()
        self.toggle_secondary_prompts()
        self.initialized = True
        self.thread.start()
        self._set_chatbot_mood()

    def set_form_values(self, _data=None):
        self.ui.prompt.blockSignals(True)
        self.ui.negative_prompt.blockSignals(True)
        self.ui.secondary_prompt.blockSignals(True)
        self.ui.secondary_negative_prompt.blockSignals(True)
        self.ui.crops_coord_top_left_x.blockSignals(True)
        self.ui.crops_coord_top_left_y.blockSignals(True)
        self.ui.image_presets.blockSignals(True)
        self.ui.quality_effects.blockSignals(True)

        self.ui.prompt.setPlainText(self.generator_settings.prompt)
        self.ui.negative_prompt.setPlainText(self.generator_settings.negative_prompt)
        self.ui.secondary_prompt.setPlainText(self.generator_settings.second_prompt)
        self.ui.secondary_negative_prompt.setPlainText(self.generator_settings.second_negative_prompt)
        self.ui.crops_coord_top_left_x.setText(str(self.generator_settings.crops_coord_top_left["x"]))
        self.ui.crops_coord_top_left_y.setText(str(self.generator_settings.crops_coord_top_left["y"]))

        image_presets = [""] + [preset.value for preset in ImagePreset]
        self.ui.image_presets.addItems(image_presets)
        self.ui.image_presets.setCurrentIndex(
            self.ui.image_presets.findText(self.generator_settings.image_preset)
        )

        self.ui.quality_effects.setCurrentText(self.generator_settings.quality_effects)

        self.ui.prompt.blockSignals(False)
        self.ui.negative_prompt.blockSignals(False)
        self.ui.secondary_prompt.blockSignals(False)
        self.ui.secondary_negative_prompt.blockSignals(False)
        self.ui.crops_coord_top_left_x.blockSignals(False)
        self.ui.crops_coord_top_left_y.blockSignals(False)
        self.ui.image_presets.blockSignals(False)
        self.ui.quality_effects.blockSignals(False)

    def clear_prompts(self):
        self.ui.prompt.setPlainText("")
        self.ui.negative_prompt.setPlainText("")
        self.ui.secondary_prompt.setPlainText("")
        self.ui.secondary_negative_prompt.setPlainText("")

    def stop_progress_bar(self, do_clear=False):
        progressbar = self.ui.progress_bar
        if not progressbar:
            return
        progressbar.setRange(0, 100)
        if do_clear:
            progressbar.setValue(0)
            progressbar.setFormat("")
        else:
            progressbar.setValue(100)
            progressbar.setFormat("Complete")

    def _set_keyboard_shortcuts(self):
        session = self.db_handler.get_db_session()
        generate_image_key = session.query(ShortcutKeys).filter_by(display_name="Generate Image").first()
        interrupt_key = session.query(ShortcutKeys).filter_by(display_name="Interrupt").first()
        if generate_image_key:
            self.ui.generate_button.setShortcut(generate_image_key.key)
            self.ui.generate_button.setToolTip(f"{generate_image_key.display_name} ({generate_image_key.text})")
        if interrupt_key:
            self.ui.interrupt_button.setShortcut(interrupt_key.key)
            self.ui.interrupt_button.setToolTip(f"{interrupt_key.display_name} ({interrupt_key.text})")
        session.close()



import os

from sqlalchemy import create_engine, Column, Integer, String, JSON, Boolean, Float, Text, ForeignKey, LargeBinary, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from airunner.enums import ImageGenerator, GeneratorSection, CanvasToolName, Mode
from airunner.settings import SD_DEFAULT_VAE_PATH, DEFAULT_SCHEDULER, \
    DEFAULT_BRUSH_PRIMARY_COLOR, DEFAULT_BRUSH_SECONDARY_COLOR, BASE_PATH


import datetime

Base = declarative_base()


class ApplicationSettings(Base):
    __tablename__ = 'application_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    use_cuda = Column(Boolean, default=True)
    sd_enabled = Column(Boolean, default=False)
    llm_enabled = Column(Boolean, default=False)
    tts_enabled = Column(Boolean, default=False)
    stt_enabled = Column(Boolean, default=False)
    controlnet_enabled = Column(Boolean, default=False)
    ai_mode = Column(Boolean, default=True)
    active_grid_size_lock = Column(Boolean, default=False)
    installation_path = Column(String, default="~/.local/share/airunner")
    current_layer_index = Column(Integer, default=0)
    paths_initialized = Column(Boolean, default=False)
    trust_remote_code = Column(Boolean, default=False)  # Leave this hardcoded. We will never trust remote code.
    nsfw_filter = Column(Boolean, default=True)
    resize_on_paste = Column(Boolean, default=True)
    image_to_new_layer = Column(Boolean, default=True)
    dark_mode_enabled = Column(Boolean, default=True)
    override_system_theme = Column(Boolean, default=True)
    latest_version_check = Column(Boolean, default=True)
    app_version = Column(String, default="")
    allow_online_mode = Column(Boolean, default=True)
    current_tool = Column(String, default=CanvasToolName.BRUSH.value)
    image_export_type = Column(String, default="png")
    auto_export_images = Column(Boolean, default=True)
    show_active_image_area = Column(Boolean, default=True)
    working_width = Column(Integer, default=512)
    working_height = Column(Integer, default=512)
    current_llm_generator = Column(String, default="causallm")
    current_image_generator = Column(String, default=ImageGenerator.STABLEDIFFUSION.value)
    generator_section = Column(String, default=GeneratorSection.TXT2IMG.value)
    hf_api_key_read_key = Column(String, default="")
    hf_api_key_write_key = Column(String, default="")
    civit_ai_api_key = Column(String, default="")
    is_maximized = Column(Boolean, default=False)
    pivot_point_x = Column(Integer, default=0)
    pivot_point_y = Column(Integer, default=0)
    mode = Column(String, default=Mode.IMAGE.value)
    autoload_sd = Column(Boolean, default=True)
    autoload_llm = Column(Boolean, default=False)
    show_nsfw_warning = Column(Boolean, default=True)
    run_setup_wizard = Column(Boolean, default=True)
    download_wizard_completed = Column(Boolean, default=False)
    stable_diffusion_agreement_checked = Column(Boolean, default=True)
    airunner_agreement_checked = Column(Boolean, default=True)
    user_agreement_checked = Column(Boolean, default=True)
    llama_license_agreement_checked = Column(Boolean, default=True)


class ActiveGridSettings(Base):
    __tablename__ = 'active_grid_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    enabled = Column(Boolean, default=True)
    render_border = Column(Boolean, default=True)
    render_fill = Column(Boolean, default=False)
    border_opacity = Column(Integer, default=50)
    fill_opacity = Column(Integer, default=50)
    border_color = Column(String, default="#00FF00")
    fill_color = Column(String, default="#FF0000")
    pos_x = Column(Integer, default=0)
    pos_y = Column(Integer, default=0)
    width = Column(Integer, default=512)
    height = Column(Integer, default=512)


class ControlnetSettings(Base):
    __tablename__ = 'controlnet_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    image = Column(String, nullable=True)
    generated_image = Column(String, nullable=True)
    enabled = Column(Boolean, default=False)
    use_grid_image_as_input = Column(Boolean, default=False)
    strength = Column(Integer, default=50)
    conditioning_scale = Column(Integer, default=100)
    guidance_scale = Column(Integer, default=750)
    controlnet = Column(String, default="Canny")


class ImageToImageSettings(Base):
    __tablename__ = 'image_to_image_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    image = Column(String, nullable=True)
    enabled = Column(Boolean, default=False)
    use_grid_image_as_input = Column(Boolean, default=False)


class OutpaintSettings(Base):
    __tablename__ = 'outpaint_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    image = Column(String, nullable=True)
    enabled = Column(Boolean, default=True)
    strength = Column(Integer, default=50)
    mask_blur = Column(Integer, default=0)


class DrawingPadSettings(Base):
    __tablename__ = 'drawing_pad_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    image = Column(String, nullable=True)
    mask = Column(String, nullable=True)
    enabled = Column(Boolean, default=True)
    enable_automatic_drawing = Column(Boolean, default=False)
    mask_layer_enabled = Column(Boolean, default=False)
    x_pos = Column(Integer, default=0)
    y_pos = Column(Integer, default=0)


class MetadataSettings(Base):
    __tablename__ = 'metadata_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    image_export_metadata_prompt = Column(Boolean, default=True)
    image_export_metadata_negative_prompt = Column(Boolean, default=True)
    image_export_metadata_scale = Column(Boolean, default=True)
    image_export_metadata_seed = Column(Boolean, default=True)
    image_export_metadata_steps = Column(Boolean, default=True)
    image_export_metadata_ddim_eta = Column(Boolean, default=True)
    image_export_metadata_iterations = Column(Boolean, default=True)
    image_export_metadata_samples = Column(Boolean, default=True)
    image_export_metadata_model = Column(Boolean, default=True)
    image_export_metadata_model_branch = Column(Boolean, default=True)
    image_export_metadata_scheduler = Column(Boolean, default=True)
    image_export_metadata_strength = Column(Boolean, default=True)
    image_export_metadata_clip_skip = Column(Boolean, default=True)
    image_export_metadata_version = Column(Boolean, default=True)
    image_export_metadata_lora = Column(Boolean, default=True)
    image_export_metadata_embeddings = Column(Boolean, default=True)
    image_export_metadata_timestamp = Column(Boolean, default=True)
    image_export_metadata_controlnet = Column(Boolean, default=True)
    export_metadata = Column(Boolean, default=True)
    import_metadata = Column(Boolean, default=True)


class GeneratorSettings(Base):
    __tablename__ = 'generator_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    pipeline_action = Column(String, default="txt2img")
    generator_name = Column(String, default="stablediffusion")
    quality_effects = Column(String, default="")
    image_preset = Column(String, default="")
    prompt = Column(String, default="")
    negative_prompt = Column(String, default="")
    second_prompt = Column(String, default="")
    second_negative_prompt = Column(String, default="")
    random_seed = Column(Boolean, default=True)
    model_name = Column(String, default="")
    model = Column(Integer, ForeignKey('aimodels.id'), nullable=True)
    aimodel = relationship("AIModels", back_populates="generator_settings")
    vae = Column(String, default=SD_DEFAULT_VAE_PATH)
    scheduler = Column(String, default=DEFAULT_SCHEDULER)
    variation = Column(Boolean, default=False)
    use_prompt_builder = Column(Boolean, default=False)
    version = Column(String, default="SD 1.5")
    is_preset = Column(Boolean, default=False)
    use_compel = Column(Boolean, default=True)

    steps = Column(Integer, default=20)
    ddim_eta = Column(Float, default=0.5)
    scale = Column(Integer, default=750)
    seed = Column(Integer, default=42)
    prompt_triggers = Column(String, default="")
    strength = Column(Integer, default=50)
    n_samples = Column(Integer, default=1)
    clip_skip = Column(Integer, default=0)
    crops_coord_top_left = Column(JSON, default={"x": 0, "y": 0})
    original_size = Column(JSON, default={"width": 512, "height": 512})
    target_size = Column(JSON, default={"width": 1024, "height": 1024})
    negative_original_size = Column(JSON, default={"width": 512, "height": 512})
    negative_target_size = Column(JSON, default={"width": 512, "height": 512})

    lora_scale = Column(Integer, default=100)


class LLMGeneratorSettings(Base):
    __tablename__ = 'llm_generator_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    action = Column(String, default="CHAT")
    use_tool_filter = Column(Boolean, default=False)
    seed = Column(Integer, default=0)
    random_seed = Column(Boolean, default=False)
    model_version = Column(String, default="w4ffl35/Mistral-7B-Instruct-v0.3-4bit")
    dtype = Column(String, default="4bit")
    use_gpu = Column(Boolean, default=True)
    message_type = Column(String, default="chat")
    override_parameters = Column(Boolean, default=True)
    current_chatbot = Column(Integer, default=0)
    prompt_template = Column(String, default="Mistral 7B Instruct: Default Chatbot")
    batch_size = Column(Integer, default=1)
    use_api = Column(Boolean, default=False)
    api_key = Column(String, nullable=True)
    api_model = Column(String, nullable=True)
    top_p = Column(Integer, default=900)
    min_length = Column(Integer, default=1)
    max_new_tokens = Column(Integer, default=1000)
    repetition_penalty = Column(Integer, default=100)
    do_sample = Column(Boolean, default=True)
    early_stopping = Column(Boolean, default=True)
    num_beams = Column(Integer, default=1)
    temperature = Column(Integer, default=1000)
    ngram_size = Column(Integer, default=2)
    top_k = Column(Integer, default=10)
    eta_cutoff = Column(Integer, default=10)
    sequences = Column(Integer, default=1)
    decoder_start_token_id = Column(Integer, nullable=True)
    use_cache = Column(Boolean, default=True)
    length_penalty = Column(Integer, default=100)


class TTSSettings(Base):
    __tablename__ = 'tts_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    tts_model = Column(String, default="SpeechT5")
    use_cuda = Column(Boolean, default=True)
    use_sentence_chunks = Column(Boolean, default=True)
    use_word_chunks = Column(Boolean, default=False)
    cuda_index = Column(Integer, default=0)
    word_chunks = Column(Integer, default=1)
    sentence_chunks = Column(Integer, default=1)
    model = Column(String, default="SpeechT5")


class SpeechT5Settings(Base):
    __tablename__ = 'speech_t5_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    datasets_path = Column(String, default="Matthijs/cmu-arctic-xvectors")
    processor_path = Column(String, default="microsoft/speecht5_tts")
    vocoder_path = Column(String, default="microsoft/speecht5_hifigan")
    model_path = Column(String, default="microsoft/speecht5_tts")
    rate = Column(Integer, default=100)
    pitch = Column(Integer, default=100)
    volume = Column(Integer, default=100)


class EspeakSettings(Base):
    __tablename__ = 'espeak_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    gender = Column(String, default="Male")
    voice = Column(String, default="male1")
    language = Column(String, default="en-US")
    rate = Column(Integer, default=100)
    pitch = Column(Integer, default=100)
    volume = Column(Integer, default=100)
    punctuation_mode = Column(String, default="none")


class STTSettings(Base):
    __tablename__ = 'stt_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    duration = Column(Integer, default=10)
    fs = Column(Integer, default=16000)
    channels = Column(Integer, default=1)
    volume_input_threshold = Column(Integer, default=0.08)
    silence_buffer_seconds = Column(Integer, default=1.0)
    chunk_duration = Column(Integer, default=0.03)


class Schedulers(Base):
    __tablename__ = 'schedulers'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=True)
    display_name = Column(String, nullable=True)


class BrushSettings(Base):
    __tablename__ = 'brush_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    size = Column(Integer, default=75)
    primary_color = Column(String, default=DEFAULT_BRUSH_PRIMARY_COLOR)
    secondary_color = Column(String, default=DEFAULT_BRUSH_SECONDARY_COLOR)
    strength_slider = Column(Integer, default=950)
    strength = Column(Integer, default=950)
    conditioning_scale = Column(Integer, default=550)
    guidance_scale = Column(Integer, default=75)


class GridSettings(Base):
    __tablename__ = 'grid_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    cell_size = Column(Integer, default=64)
    line_width = Column(Integer, default=1)
    line_color = Column(String, default="#101010")
    snap_to_grid = Column(Boolean, default=True)
    canvas_color = Column(String, default="#000000")
    show_grid = Column(Boolean, default=True)
    zoom_level = Column(Float, default=1.0)
    zoom_in_step = Column(Float, default=0.1)
    zoom_out_step = Column(Float, default=0.1)


class PathSettings(Base):
    __tablename__ = 'path_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    base_path = Column(String, default=BASE_PATH)
    documents_path = Column(String, default=os.path.expanduser(os.path.join(BASE_PATH, "text/other", "documents")))
    ebook_path = Column(String, default=os.path.expanduser(os.path.join(BASE_PATH, "text/other", "ebooks")))
    image_path = Column(String, default=os.path.expanduser(os.path.join(BASE_PATH, "art/other", "images")))
    llama_index_path = Column(String, default=os.path.expanduser(os.path.join(BASE_PATH, "text/rag", "db")))
    webpages_path = Column(String, default=os.path.expanduser(os.path.join(BASE_PATH, "text/other", "webpages")))
    stt_model_path = Column(String, default=os.path.expanduser(os.path.join(BASE_PATH, "text/models/stt", "models")))
    tts_model_path = Column(String, default=os.path.expanduser(os.path.join(BASE_PATH, "text/models/tts", "models")))


class MemorySettings(Base):
    __tablename__ = 'memory_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    use_last_channels = Column(Boolean, default=True)
    use_attention_slicing = Column(Boolean, default=False)
    use_tf32 = Column(Boolean, default=False)
    use_enable_vae_slicing = Column(Boolean, default=True)
    use_accelerated_transformers = Column(Boolean, default=True)
    use_tiled_vae = Column(Boolean, default=True)
    enable_model_cpu_offload = Column(Boolean, default=False)
    use_enable_sequential_cpu_offload = Column(Boolean, default=False)
    use_cudnn_benchmark = Column(Boolean, default=True)
    use_torch_compile = Column(Boolean, default=False)
    use_tome_sd = Column(Boolean, default=True)
    tome_sd_ratio = Column(Integer, default=600)
    move_unused_model_to_cpu = Column(Boolean, default=False)
    unload_unused_models = Column(Boolean, default=True)
    default_gpu_sd = Column(Integer, default=0)
    default_gpu_llm = Column(Integer, default=0)
    default_gpu_tts = Column(Integer, default=0)
    default_gpu_stt = Column(Integer, default=0)


class Chatbot(Base):
    __tablename__ = 'chatbots'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, default="Chatbot")
    username = Column(String, default="User")
    botname = Column(String, default="Computer")
    use_personality = Column(Boolean, default=True)
    use_mood = Column(Boolean, default=True)
    use_guardrails = Column(Boolean, default=True)
    use_system_instructions = Column(Boolean, default=True)
    use_datetime = Column(Boolean, default=True)
    assign_names = Column(Boolean, default=True)
    bot_personality = Column(Text, default="happy. He loves {{ username }}")
    bot_mood = Column(Text, default="")
    prompt_template = Column(Text, default="Mistral 7B Instruct: Default Chatbot")
    use_tool_filter = Column(Boolean, default=False)
    use_gpu = Column(Boolean, default=True)
    skip_special_tokens = Column(Boolean, default=True)
    sequences = Column(Integer, default=1)
    seed = Column(Integer, default=42)
    random_seed = Column(Boolean, default=True)
    model_version = Column(String, default="w4ffl35/Mistral-7B-Instruct-v0.3-4bit")
    model_type = Column(String, default="llm")
    dtype = Column(String, default="4bit")
    return_result = Column(Boolean, default=True)
    guardrails_prompt = Column(Text, default=(
        "Always assist with care, respect, and truth. "
        "Respond with utmost utility yet securely. "
        "Avoid harmful, unethical, prejudiced, or negative content. "
        "Ensure replies promote fairness and positivity."
    ))
    system_instructions = Column(Text, default=(
        "You are a dialogue generator. "
        "You will follow all of the rules in order to generate compelling and intriguing dialogue for a given character.\n"
        "The Rules:\n"
        "You will ONLY return dialogue, nothing more.\n"
        "Limit responses to a single sentence.\n"
        "Only generate responses in pure dialogue form without including any actions, descriptions or stage directions in parentheses. Only return spoken words.\n"
        "Do not generate redundant dialogue. Examine the conversation and context close and keep responses interesting and creative.\n"
        "Do not format the response with the character's name or any other text. Only return the dialogue.\n"
        "Respond with dialogue that is appropriate for a character named {{ speaker_name }}.\n"
        "{{ speaker_name }} and {{ listener_name }} are having a conversation. \n"
        "Avoid repeating {{ speaker_name }}'s previous dialogue or {{ listener_name }}'s previous dialogue.\n"
        "You will generate responses which are appropriate for your personality and given character.\n"
        "------\n"
    ))
    top_p = Column(Integer, default=900)
    min_length = Column(Integer, default=1)
    max_new_tokens = Column(Integer, default=1000)
    repetition_penalty = Column(Integer, default=100)
    do_sample = Column(Boolean, default=True)
    early_stopping = Column(Boolean, default=True)
    num_beams = Column(Integer, default=1)
    temperature = Column(Integer, default=1000)
    ngram_size = Column(Integer, default=2)
    top_k = Column(Integer, default=10)
    eta_cutoff = Column(Integer, default=10)
    num_return_sequences = Column(Integer, default=1)
    decoder_start_token_id = Column(Integer, default=None)
    use_cache = Column(Boolean, default=True)
    length_penalty = Column(Integer, default=100)
    seed = Column(Integer, default=0)
    random_seed = Column(Boolean, default=False)

    target_files = relationship("TargetFiles", back_populates="chatbot")
    target_directories = relationship("TargetDirectories", back_populates="chatbot")
    messages = relationship("Message", back_populates="chatbot")


class TargetFiles(Base):
    __tablename__ = 'target_files'
    id = Column(Integer, primary_key=True, autoincrement=True)
    chatbot_id = Column(Integer, ForeignKey('chatbots.id'))
    file_path = Column(String)

    chatbot = relationship("Chatbot", back_populates="target_files")


class TargetDirectories(Base):
    __tablename__ = 'target_directories'
    id = Column(Integer, primary_key=True, autoincrement=True)
    chatbot_id = Column(Integer, ForeignKey('chatbots.id'))
    directory_path = Column(String)

    chatbot = relationship("Chatbot", back_populates="target_directories")


class AIModels(Base):
    __tablename__ = 'aimodels'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    path = Column(String, nullable=False)
    branch = Column(String, nullable=False)
    version = Column(String, nullable=False)
    category = Column(String, nullable=False)
    pipeline_action = Column(String, nullable=False)
    enabled = Column(Boolean, nullable=False)
    model_type = Column(String, nullable=False)
    is_default = Column(Boolean, nullable=False)
    generator_settings = relationship("GeneratorSettings", back_populates="aimodel")


class ShortcutKeys(Base):
    __tablename__ = 'shortcut_keys'
    id = Column(Integer, primary_key=True, autoincrement=True)
    display_name = Column(String, nullable=False)
    text = Column(String, nullable=False)
    key = Column(Integer, nullable=False)
    modifiers = Column(Integer, nullable=False)
    description = Column(String, nullable=False)
    signal = Column(Integer, nullable=False)


class Lora(Base):
    __tablename__ = 'lora'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    scale = Column(Integer, nullable=False)
    enabled = Column(Boolean, nullable=False)
    loaded = Column(Boolean, default=False, nullable=False)
    trigger_word = Column(String, nullable=True)
    path = Column(String, nullable=True)
    version = Column(String, nullable=True)


class SavedPrompt(Base):
    __tablename__ = "saved_prompts"
    id = Column(Integer, primary_key=True, autoincrement=True)
    prompt = Column(String, nullable=True)
    secondary_prompt = Column(String, nullable=True)
    negative_prompt = Column(String, nullable=True)
    secondary_negative_prompt = Column(String, nullable=True)


class Embedding(Base):
    __tablename__ = "embeddings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    path = Column(String, nullable=False)
    version = Column(String, nullable=False)
    tags = Column(String, default="")
    active = Column(Boolean, default=False)
    trigger_word = Column(String, default="")


class PromptTemplate(Base):
    __tablename__ = "prompt_templates"
    id = Column(Integer, primary_key=True, autoincrement=True)
    template_name = Column(String, nullable=False)
    use_guardrails = Column(Boolean, default=True)
    guardrails = Column(Text, default="")
    system = Column(Text, default="")
    use_system_datetime_in_system_prompt = Column(Boolean, default=False)


class ControlnetModel(Base):
    __tablename__ = "controlnet_models"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    path = Column(String, nullable=False)
    display_name = Column(String, nullable=False)
    version = Column(String, nullable=False)


class FontSetting(Base):
    __tablename__ = "font_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    font_family = Column(String, nullable=False)
    font_size = Column(Integer, nullable=False)


class PipelineModel(Base):
    __tablename__ = "pipeline_models"
    id = Column(Integer, primary_key=True, autoincrement=True)
    pipeline_action = Column(String, nullable=False)
    version = Column(String, nullable=False)
    category = Column(String, nullable=False)
    classname = Column(String, nullable=False)
    default = Column(Boolean, nullable=False)


class WindowSettings(Base):
    __tablename__ = "window_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    is_maximized = Column(Boolean, default=False)
    is_fullscreen = Column(Boolean, default=False)
    llm_splitter = Column(LargeBinary, nullable=True)
    content_splitter = Column(LargeBinary, nullable=True)
    generator_form_splitter = Column(LargeBinary, nullable=True)
    grid_settings_splitter = Column(LargeBinary, nullable=True)
    tool_tab_widget_index = Column(Integer, default=0)
    width = Column(Integer, default=800)
    height = Column(Integer, default=600)
    x_pos = Column(Integer, default=0)
    y_pos = Column(Integer, default=0)
    mode_tab_widget_index = Column(Integer, default=0)


class Conversation(Base):
    __tablename__ = 'conversations'
    id = Column(Integer, primary_key=True, autoincrement=True)
    timestamp = Column(DateTime, default=datetime.datetime.now(datetime.timezone.utc))
    title = Column(String, nullable=True)  # New column added
    messages = relationship("Message", back_populates="conversation", cascade="all, delete-orphan")


class Message(Base):
    __tablename__ = 'messages'
    id = Column(Integer, primary_key=True, autoincrement=True)
    role = Column(String, nullable=False)
    content = Column(String, nullable=False)
    timestamp = Column(DateTime, default=datetime.datetime.now(datetime.timezone.utc))
    conversation_id = Column(Integer, ForeignKey('conversations.id'))
    conversation = relationship("Conversation", back_populates="messages")
    name = Column(String, nullable=True)  # New column added
    is_bot = Column(Boolean, default=False)  # New column added
    chatbot_id = Column(Integer, ForeignKey('chatbots.id'))

    chatbot = relationship("Chatbot", back_populates="messages")


class Summary(Base):
    __tablename__ = 'summaries'
    id = Column(Integer, primary_key=True, autoincrement=True)
    content = Column(String, nullable=False)
    timestamp = Column(DateTime, default=datetime.datetime.now(datetime.timezone.utc))
    conversation_id = Column(Integer, ForeignKey('conversations.id'))
    conversation = relationship("Conversation", back_populates="summaries")


Conversation.messages = relationship("Message", order_by=Message.id, back_populates="conversation")
Conversation.summaries = relationship("Summary", order_by=Summary.id, back_populates="conversation")
Message.chatbot = relationship("Chatbot", back_populates="messages")


class ImageFilter(Base):
    __tablename__ = 'image_filter_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    display_name = Column(String, nullable=False)
    auto_apply = Column(Boolean, default=False)
    filter_class = Column(String, nullable=False)


class ImageFilterValue(Base):
    __tablename__ = 'image_filter_values'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    value = Column(String, nullable=False)
    value_type = Column(String, nullable=False)
    min_value = Column(Float, nullable=True)
    max_value = Column(Float, nullable=True)
    image_filter_id = Column(Integer, ForeignKey('image_filter_settings.id'))
    image_filter = relationship("ImageFilter", back_populates="image_filter_values")

ImageFilter.image_filter_values = relationship("ImageFilterValue", order_by=ImageFilterValue.id, back_populates="image_filter")
ImageFilterValue.image_filter = relationship("ImageFilter", back_populates="image_filter_values")



from typing import Optional

from PySide6.QtCore import Qt, QPoint

from airunner.cursors.circle_brush import CircleCursor
from airunner.enums import SignalCode, CanvasToolName
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.canvas.templates.canvas_ui import Ui_canvas


class CanvasWidget(
    BaseWidget
):
    """
    Widget responsible for multiple functionalities:

    - Allows the user to draw on a canvas.
    - Displays the grid.
    - Displays images.
    - Handles the active grid area.
    """
    widget_class_ = Ui_canvas

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._startPos = QPoint(0, 0)
        self.images = {}
        self.active_grid_area_pivot_point = QPoint(0, 0)
        self.active_grid_area_position = QPoint(0, 0)
        self.current_image_index = 0
        self.draggable_pixmaps_in_scene = {}
        self.drag_pos: QPoint = Optional[None]
        self._grid_settings = {}
        self._active_grid_settings = {}
        self.signal_handlers = {
            SignalCode.CANVAS_UPDATE_CURSOR: self.on_canvas_update_cursor_signal,
        }

    @property
    def image_pivot_point(self):
        settings = self.application_settings
        try:
            return QPoint(
                settings.pivot_point_x,
                settings.pivot_point_y
            )
        except Exception as e:
            self.logger.error(e)
        return QPoint(0, 0)

    @image_pivot_point.setter
    def image_pivot_point(self, value):
        settings = self.application_settings
        settings.pivot_point_x = value.x()
        settings.pivot_point_y = value.y()
        self.update_application_settings("pivot_point_x", value.x())
        self.update_application_settings("pivot_point_y", value.y())

    def on_canvas_update_cursor_signal(self, message: dict):
        event = message.get("event", None)
        if self.current_tool in (
            CanvasToolName.BRUSH,
            CanvasToolName.ERASER
        ):
            cursor = CircleCursor(
                Qt.GlobalColor.white,
                Qt.GlobalColor.transparent,
                self.brush_settings.size,
            )
        elif self.current_tool is CanvasToolName.ACTIVE_GRID_AREA:
            if event and event.buttons() == Qt.MouseButton.LeftButton:
                cursor = Qt.CursorShape.ClosedHandCursor
            else:
                cursor = Qt.CursorShape.OpenHandCursor
        else:
            cursor = Qt.CursorShape.ArrowCursor
        self.setCursor(cursor)

    def toggle_grid(self, val):
        self.do_draw()

    def showEvent(self, event):
        super().showEvent(event)
        self.do_draw(force_draw=True)

    def do_draw(
        self,
        force_draw: bool = False
    ):
        self.emit_signal(SignalCode.SCENE_DO_DRAW_SIGNAL, {
            "force_draw": force_draw
        })
        self.ui.canvas_container_size = self.ui.canvas_container.viewport().size()



import os
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog
from airunner.mediator_mixin import MediatorMixin
from airunner.settings import DARK_THEME_NAME, LIGHT_THEME_NAME
from airunner.windows.main.ai_model_mixin import AIModelMixin
from airunner.windows.main.settings_mixin import SettingsMixin


class BaseWindow(
    QDialog,
    MediatorMixin,
    SettingsMixin,
    AIModelMixin
):
    template_class_ = None
    template = None
    is_modal: bool = False  # allow the window to be treated as a modal
    title: str = "Base Window"

    def __init__(self, **kwargs):
        MediatorMixin.__init__(self)
        SettingsMixin.__init__(self)
        AIModelMixin.__init__(self)
        super().__init__()
        self.do_exec = kwargs.get("exec", True)

        self.set_stylesheet()

        self.ui = self.template_class_()
        self.ui.setupUi(self)
        if self.is_modal:
            self.setWindowModality(Qt.WindowModality.WindowModal)
            self.setWindowFlag(Qt.WindowType.WindowStaysOnTopHint)
        self.setWindowTitle(self.title)
        self.initialize_window()
        if self.do_exec:
            self.exec()

    def initialize_window(self):
        pass

    def set_stylesheet(self):
        """
        Sets the stylesheet for the application based on the current theme
        """
        if self.application_settings.override_system_theme:
            theme_name = DARK_THEME_NAME if self.application_settings.dark_mode_enabled else LIGHT_THEME_NAME
            here = os.path.dirname(os.path.realpath(__file__))
            with open(os.path.join(here, "..", "styles", theme_name, "styles.qss"), "r") as f:
                stylesheet = f.read()
            self.setStyleSheet(stylesheet)
        else:
            self.setStyleSheet("")
        self.update()



####################################################################################################
# Do not remove the following import statement or change
# the order of the imports.
####################################################################################################
import sys
import signal

from PySide6.QtCore import (
    QObject
)
from PySide6.QtGui import (
    Qt
)
from PySide6.QtWidgets import (
    QApplication
)

from airunner.mediator_mixin import MediatorMixin
from airunner.windows.download_wizard.download_wizard_window import DownloadWizardWindow
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.windows.setup_wizard.setup_wizard_window import SetupWizardWindow
from airunner.aihandler.logger import Logger


class AppInstaller(
    QObject,
    SettingsMixin,
    MediatorMixin
):
    """
    The main application class for AI Runner.
    This class can be run as a GUI application or as a socket server.
    """
    def __init__(
        self,
        close_on_cancel:bool = True
    ):
        """
        Initialize the application and run as a GUI application or a socket server.
        """

        self.wizard = None
        self.download_wizard = None
        self.app = None
        self.close_on_cancel = close_on_cancel
        self.logger = Logger(prefix=self.__class__.__name__)

        """
        Mediator and Settings mixins are initialized here, enabling the application
        to easily access the application settings dictionary.
        """
        MediatorMixin.__init__(self)
        SettingsMixin.__init__(self)
        super(AppInstaller, self).__init__()

        self.start()

    @property
    def do_show_setup_wizard(self) -> bool:
        """
        This flag is used to determine if the setup wizard should be displayed.
        If the setup wizard has not been dcompleted, various agreements have not been accepted,
        or the paths have not been initialized, the setup wizard will be displayed.
        :return: bool
        """
        return (
            (
                # self.wizard.setup_settings["paths_initialized"] and
                self.application_settings.user_agreement_checked and
                self.application_settings.stable_diffusion_agreement_checked and
                self.application_settings.airunner_agreement_checked
            )
        )

    def start(self):
        """
        Conditionally initialize and display the setup wizard.
        :return:
        """
        signal.signal(signal.SIGINT, self.signal_handler)
        QApplication.setAttribute(Qt.ApplicationAttribute.AA_UseDesktopOpenGL)
        self.app = QApplication.instance()

        self.wizard = SetupWizardWindow()
        self.wizard.exec()

        if self.wizard.canceled:
            print("canceled")
            self.cancel()
            return

        self.download_wizard = DownloadWizardWindow()
        self.download_wizard.exec()

    @staticmethod
    def signal_handler(
        _signal,
        _frame
    ):
        """
        Handle the SIGINT signal in a clean way.
        :param _signal:
        :param _frame:
        :return: None
        """
        print("\nExiting...")
        try:
            AppInstaller.quit()
            sys.exit(0)
        except Exception as e:
            print(e)
            sys.exit(0)

    def cancel(self):
        self.wizard.close()
        if self.download_wizard:
            self.download_wizard.close()
        if self.close_on_cancel:
            self.quit()
            sys.exit(0)

    @staticmethod
    def quit():
        app = QApplication.instance()
        app.quit()



class BaseMixin:
    def __init__(self, data):
        self.data = data

    def get_all(self):
        return self.data

    def get_by_filter(self, filter_dict):
        return [item for item in self.data if all(item.get(k) == v for k, v in filter_dict.items())]

    def create(self, item):
        self.data.append(item)

    def update(self, item):
        for i, existing_item in enumerate(self.data):
            if existing_item['name'] == item['name']:
                self.data[i] = item
                break

    def delete(self, item):
        self.data = [existing_item for existing_item in self.data if existing_item['name'] != item['name']]



import random


class BuildPrompt:
    @classmethod
    def process_variable(cls, var):
        if isinstance(var, dict):
            if "range" in var:
                if "type" in var and var["type"] == "range":
                    var = random.randint(var["min"], var["max"])
        return var

    @classmethod
    def has_variable(cls, variable, available_variables):
        return variable in available_variables and available_variables[variable] != [""]

    @classmethod
    def add_text_to_prompt(
        cls,
        prompt,
        text,
        cond,
        cond_val,
        not_cond,
        not_cond_val,
        or_cond,
        or_cond_val,
        else_cond,
        else_cond_val
    ):
        cond_true = (cond and cond_val) or (not cond)
        not_cond_true = (not_cond and not not_cond_val) or (not not_cond)
        or_cond_true = (or_cond and or_cond_val) or (not or_cond)

        if cond_true and not_cond_true and or_cond_true:
            if text:
                return prompt + text, True
            else:
                return prompt, True
        elif else_cond and else_cond_val:
            return prompt + else_cond, True
        return prompt, False

    @classmethod
    def build_conditional_prompt(cls, conditionals, vars, category, prompt=""):
        for conditional in conditionals:
            text = None
            cond = None
            not_cond = None
            or_cond = None
            next = []
            else_cond = None
            if "text" in conditional:
                text = conditional["text"]
            if "cond" in conditional:
                cond = conditional["cond"]
            if "not_cond" in conditional:
                not_cond = conditional["not_cond"]
            if "next" in conditional:
                next = conditional["next"]
            if "else" in conditional:
                else_cond = conditional["else"]
            if "or_cond" in conditional:
                or_cond = conditional["or_cond"]

            if not text and len(next) == 0:
                continue

            if cond:
                has_cond = True
                if isinstance(cond, list):
                    for cond_var in cond:
                        if not cls.has_variable(cond_var, vars):
                            has_cond = False
                else:
                    if not cls.has_variable(cond, vars):
                        has_cond = False
            else:
                has_cond = False

            if not_cond:
                not_cond_val = False
                if isinstance(not_cond, list):
                    for not_cond_var in not_cond:
                        if cls.has_variable(not_cond_var, vars):
                            not_cond_val = True
                else:
                    if cls.has_variable(not_cond, vars):
                        not_cond_val = True
            else:
                not_cond_val = True

            else_cond_val = True if else_cond else False

            or_cond_val = False
            if or_cond:
                if isinstance(or_cond, list):
                    for or_cond_var in or_cond:
                        if cls.has_variable(or_cond_var, vars):
                            or_cond_val = True
                            break
                else:
                    if cls.has_variable(or_cond, vars):
                        or_cond_val = True

            prompt, success = cls.add_text_to_prompt(
                prompt,
                text,
                cond,
                has_cond,
                not_cond,
                not_cond_val,
                or_cond,
                or_cond_val,
                else_cond,
                else_cond_val
            )

            if success:
                for next_conditionals in next:
                    prompt = cls.build_conditional_prompt(
                        next_conditionals,
                        vars,
                        category,
                        prompt
                    )

        return prompt

    @classmethod
    def build_prompt(cls, conditionals, vars, category):
        prompt = cls.build_conditional_prompt(
            conditionals,
            vars,
            category)
        return prompt.strip()



import json
import re


class JSONExtractor(json.JSONDecoder):
    def decode(self, s):
        self.json_objects = []
        # Remove trailing characters
        while s:
            try:
                # Try to decode the string as JSON
                obj, end = super().raw_decode(s)
                self.json_objects.append(obj)
                s = s[end:].lstrip()
            except json.JSONDecodeError:
                # If the string is not valid JSON, find the first valid JSON object
                match = re.match(r'[^\{]+(\{[^\}]+\}).*', s)
                if match:
                    json_part = match.group(1)  # Extract the JSON object from the match
                    s = s[match.end():]
                    s = s.replace("```json", "").replace("```", "")
                    try:
                        # Try to decode the extracted part as JSON
                        obj = json.loads(json_part)
                        self.json_objects.append(obj)
                    except json.JSONDecodeError:
                        # If the extracted part is not valid JSON, ignore it
                        pass
                else:
                    break
        return self.json_objects

    def raw_decode(self, s, idx=0):
        obj, end = super().raw_decode(s, idx)
        self.json_objects.append(obj)
        return obj, end



from PySide6.QtCore import Qt
from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import QSpacerItem, QSizePolicy, QWidget

from airunner.aihandler.models.settings_models import ShortcutKeys
from airunner.enums import SignalCode
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.keyboard_shortcuts.templates.keyboard_shortcut_widget_ui import Ui_keyboard_shortcut_widget
from airunner.widgets.keyboard_shortcuts.templates.keyboard_shortcuts_ui import Ui_keyboard_shortcuts


class KeyboardShortcutsWidget(BaseWidget):
    widget_class_ = Ui_keyboard_shortcuts

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.spacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.shortcut_key_widgets = [None for _i in range(len(self.shortcut_keys))]
        self.pressed_keys = set()

    def showEvent(self, event):
        super().showEvent(event)
        self.initialize_ui()

    def initialize_ui(self):
        for index, shortcut_key in enumerate(self.shortcut_keys):
            self.add_widget(index, shortcut_key)
        self.ui.scrollAreaWidgetContents.layout().addItem(self.spacer)

    def add_widget(self, index: int, shortcut_key: ShortcutKeys):
        widget = QWidget()
        ui = Ui_keyboard_shortcut_widget()
        ui.setupUi(widget)
        ui.label.setText(shortcut_key.display_name)
        ui.line_edit.setText(shortcut_key.text)
        ui.line_edit.mousePressEvent = lambda event: self.set_shortcut(index, ui.line_edit)
        ui.line_edit.keyReleaseEvent = lambda event: self.get_shortcut(shortcut_key, ui.line_edit, event, index)
        self.ui.scrollAreaWidgetContents.layout().addWidget(widget)
        self.shortcut_key_widgets[index] = ui
        return widget

    def set_shortcut(self, key, line_edit):
        self.clear_shortcut_setting(key)
        line_edit.setText("Press any key to set shortcut (esc to cancel)")

    def get_shortcut(self, shortcut_key: ShortcutKeys, line_edit, event, index):
        if event.isAutoRepeat():
            return

        if event.type() == QtCore.QEvent.Type.KeyPress:
            self.pressed_keys.add(event.key())
        elif event.type() == QtCore.QEvent.Type.KeyRelease:
            self.pressed_keys.discard(event.key())

        # Check if all keys are released
        if not self.pressed_keys:
            if event.key() in [
                QtCore.Qt.Key.Key_Control,
                QtCore.Qt.Key.Key_Shift,
                QtCore.Qt.Key.Key_Alt,
                QtCore.Qt.Key.Key_Meta
            ]:
                return
            if event.key() == Qt.Key.Key_Escape:
                line_edit.setText(shortcut_key.text)
                return

            shortcut_key.text = self.get_key_text(event)
            shortcut_key.key = event.key()
            shortcut_key.modifiers = event.modifiers().value

            session = self.db_handler.get_db_session()
            session.add(shortcut_key)


            # clear existing key if it exists
            existing_keys = session.query(ShortcutKeys).filter(
                ShortcutKeys.text == shortcut_key.text,
                ShortcutKeys.id != shortcut_key.id
            ).all()
            for existing_key in existing_keys:
                existing_key.text = ""
                existing_key.key = 0
                existing_key.modifiers = 0
                session.add(existing_key)

            for i, widget in enumerate(self.shortcut_key_widgets):
                if i == index:
                    continue
                if widget.line_edit.text() == shortcut_key.text:
                    widget.line_edit.setText("")

            line_edit.setText(shortcut_key.text)
            session.commit()
            session.close()

            self.pressed_keys.clear()
            self.emit_signal(SignalCode.KEYBOARD_SHORTCUTS_UPDATED)

    def get_key_text(self, event):
        key_sequence = QtGui.QKeySequence(event.key() | event.modifiers().value)
        return key_sequence.toString(QtGui.QKeySequence.SequenceFormat.NativeText)

    def save_shortcuts(self):
        session = self.db_handler.get_db_session()
        for k, v in enumerate(self.shortcut_keys):
            # Ensure v.modifiers is a list
            if not isinstance(v.modifiers, list):
                v.modifiers = []
            session.query(ShortcutKeys).filter(ShortcutKeys.id == v.id).update({
                "text": v.text,
                "key": v.key,
                "modifiers": ",".join(v.modifiers)  # Convert list to comma-separated string
            })
        session.commit()
        session.close()

    def clear_shortcut_setting(self, key=""):
        for index, v in enumerate(self.shortcut_keys):
            if v.key != key:
                self.shortcut_key_widgets[index].line_edit.setText(v.text)



"""
====================================================================
--------------------------------------------------------------------
CAPSIZE LLC
--------------------------------------------------------------------
====================================================================

See the user agreement and license agreement and other applicable
agreements before using this software.

If you find any application which uses AI Runner as a backend,
audit the application by using this repository as a reference.

All current and future security documentation will be contained
within this file.
"""
import logging
import os
import sys
from PySide6 import QtCore
from airunner.enums import (
    Scheduler,
    SignalCode,
    Gender,
)
python_venv_dir = os.path.dirname(sys.executable)
NLTK_DOWNLOAD_DIR = os.path.join(
    python_venv_dir,
    "..",
    "lib/python3.10/site-packages/llama_index/legacy/_static/nltk_cache/"
)
ORGANIZATION = "Capsize Games"
APPLICATION_NAME = "AI Runner"
LOG_LEVEL = logging.DEBUG
DEFAULT_LLM_HF_PATH = "w4ffl35/Mistral-7B-Instruct-v0.3-4bit"
DEFAULT_STT_HF_PATH = "openai/whisper-tiny"
DEFAULT_IMAGE_SYSTEM_PROMPT = "\n".join([
    (
        "You are an image captioning expert. You will be given the "
        "description of an image. Your goal is to convert that "
        "description into a better, more fitting description which "
        "will capture the essence and the details of the image. "
        "Use parentheses to indicate the most important details of the "
        "image. Add a plus sign after a word or parenthesis to add "
        "extra emphasis. More plus signs indicate more emphasis. Minus "
        "signs can be used to indicate less emphasis."
        "You should describe the image type (professional photograph, "
        "portrait, illustration etc)"
        "You should also describe the lighting (well-lit, dim, "
        "dark etc), the color, the composition and the mood."
    ),
])
DEFAULT_IMAGE_LLM_GUARDRAILS = (
    "Avoid generating images that are illegal, "
    "harmful, or might be seen as offensive."
)
BASE_PATH = "~/.local/share/airunner"
DEFAULT_PATH_SETTINGS = {
    "base_path": BASE_PATH,
    "documents_path": os.path.expanduser(
        os.path.join(
            BASE_PATH,
            "text/other",
            "documents"
        )
    ),
    "ebook_path": os.path.expanduser(
        os.path.join(
            BASE_PATH,
            "text/other",
            "ebooks"
        )
    ),
    "image_path": os.path.expanduser(
        os.path.join(
            BASE_PATH,
            "art/other",
            "images"
        )
    ),
    "llama_index_path": os.path.expanduser(
        os.path.join(
            BASE_PATH,
            "text/rag",
            "db"
        )
    ),
    "webpages_path": os.path.expanduser(
        os.path.join(
            BASE_PATH,
            "text/other",
            "webpages"
        )
    ),
}
MALE = Gender.MALE
FEMALE = Gender.FEMALE
PHOTO_REALISTIC_NEGATIVE_PROMPT = (
    "illustration, drawing, cartoon, not real, fake, cgi, 3d animation, "
    "3d art, sculpture, animation, anime, Digital art, Concept art, Pixel art"
)

ILLUSTRATION_NEGATIVE_PROMPT = (
    "photo, photograph, photography, high-definition, video, "
    "realistic, hyper-realistic, film"
)
BUG_REPORT_LINK = (
    "https://github.com/Capsize-Games/airunner/issues/new"
    "?assignees=&labels=&template=bug_report.md&title="
)
VULNERABILITY_REPORT_LINK = (
    "https://github.com/Capsize-Games/airunner/security/advisories/new"
)
SD_DEFAULT_VAE_PATH = ""
SD_FEATURE_EXTRACTOR_PATH = "openai/clip-vit-large-patch14"
DEFAULT_BRUSH_PRIMARY_COLOR = "#99C1F1"
DEFAULT_BRUSH_SECONDARY_COLOR = "#000000"
STATUS_ERROR_COLOR = "#ff0000"
STATUS_NORMAL_COLOR_LIGHT = "#000000"
STATUS_NORMAL_COLOR_DARK = "#ffffff"
DARK_THEME_NAME = "dark_theme"
LIGHT_THEME_NAME = "light_theme"
VALID_IMAGE_FILES = "Image Files (*.png *.jpg *.jpeg)"
ESPEAK_SETTINGS = {
    "voices": {
        "Male": [
            "m1", "m2", "m3",
        ],
        "Female": [
            "f1", "f2", "f3",
        ],
    },
    "rate": {
        "min": -100,
        "max": 100,
        "default": 0
    },
    "pitch": {
        "min": -100,
        "max": 100,
        "default": 0
    },
    "volume": {
        "min": 0,
        "max": 100,
        "default": 100
    },
    "punctuation_modes": ["none", "all", "some"],
}
AVAILABLE_ACTIONS = [
    "txt2img",
    "img2img",
    "outpaint",
    "controlnet",
    "safety_checker",
]
SCHEDULER_CLASSES = [
    dict(
        display_name=Scheduler.EULER_ANCESTRAL.value,
        name="EulerAncestralDiscreteScheduler",
    ),
    dict(
        display_name=Scheduler.EULER.value,
        name="EulerDiscreteScheduler",
    ),
    dict(
        display_name=Scheduler.LMS.value,
        name="LMSDiscreteScheduler",
    ),
    dict(
        display_name=Scheduler.HEUN.value,
        name="HeunDiscreteScheduler",
    ),
    dict(
        display_name=Scheduler.DPM2.value,
        name="DPMSolverSinglestepScheduler",
    ),
    dict(
        display_name=Scheduler.DPM_PP_2M.value,
        name="DPMSolverMultistepScheduler",
    ),
    dict(
        display_name=Scheduler.DPM2_K.value,
        name="KDPM2DiscreteScheduler",
    ),
    dict(
        display_name=Scheduler.DPM2_A_K.value,
        name="KDPM2AncestralDiscreteScheduler",
    ),
    dict(
        display_name=Scheduler.DPM_PP_2M_K.value,
        name="DPMSolverMultistepScheduler",
    ),
    dict(
        display_name=Scheduler.DPM_PP_2M_SDE_K.value,
        name="DPMSolverMultistepScheduler",
    ),
    dict(
        display_name=Scheduler.DDIM.value,
        name="DDIMScheduler",
    ),
    dict(
        display_name=Scheduler.UNIPC.value,
        name="UniPCMultistepScheduler",
    ),
    dict(
        display_name=Scheduler.DDPM.value,
        name="DDPMScheduler",
    ),
    dict(
        display_name=Scheduler.DEIS.value,
        name="DEISMultistepScheduler",
    ),
    dict(
        display_name=Scheduler.DPM_2M_SDE_K.value,
        name="DPMSolverMultistepScheduler",
    ),
    dict(
        display_name=Scheduler.PLMS.value,
        name="PNDMScheduler",
    ),
    dict(
        display_name=Scheduler.DPM.value,
        name="DPMSolverMultistepScheduler",
    ),
]
MAX_SEED = 4294967295
SCHEDULERS = [e.value for e in Scheduler]
DEFAULT_SCHEDULER = Scheduler.DPM_PP_2M_K.value
MIN_NUM_INFERENCE_STEPS_IMG2IMG = 3
NSFW_CONTENT_DETECTED_MESSAGE = "NSFW content detected"
SLEEP_TIME_IN_MS = 50
DEFAULT_SHORTCUTS = [
    {
        "display_name": "Generate Image",
        "text": "F1",
        "key": QtCore.Qt.Key.Key_F1.value,
        "modifiers": QtCore.Qt.KeyboardModifier.NoModifier.value,
        "description": "Generate key. Responsible for triggering the generation of a Stable Diffusion image.",
        "signal": SignalCode.SD_GENERATE_IMAGE_SIGNAL.value
    },
    {
        "display_name": "Brush Tool",
        "text": "B",
        "key": QtCore.Qt.Key.Key_B.value,
        "modifiers": QtCore.Qt.KeyboardModifier.NoModifier.value,
        "description": "Brush tool key. Responsible for selecting the brush tool.",
        "signal": SignalCode.ENABLE_BRUSH_TOOL_SIGNAL.value
    },
    {
        "display_name": "Eraser Tool",
        "text": "E",
        "key": QtCore.Qt.Key.Key_E.value,
        "modifiers": QtCore.Qt.KeyboardModifier.NoModifier.value,
        "description": "Eraser tool key. Responsible for selecting the eraser tool.",
        "signal": SignalCode.ENABLE_ERASER_TOOL_SIGNAL.value
    },
    {
        "display_name": "Move Tool",
        "text": "V",
        "key": QtCore.Qt.Key.Key_V.value,
        "modifiers": QtCore.Qt.KeyboardModifier.NoModifier.value,
        "description": "Move tool key. Responsible for selecting the move tool.",
        "signal": SignalCode.ENABLE_MOVE_TOOL_SIGNAL.value
    },
    {
        "display_name": "Select Tool",
        "text": "S",
        "key": QtCore.Qt.Key.Key_S.value,
        "modifiers": QtCore.Qt.KeyboardModifier.NoModifier.value,
        "description": "Select tool key. Responsible for selecting the select tool.",
        "signal": SignalCode.ENABLE_SELECTION_TOOL_SIGNAL.value
    },
    {
        "display_name": "Interrupt",
        "text": "Shift+Ctrl+I",
        "key": QtCore.Qt.Key.Key_I.value,
        "modifiers": QtCore.Qt.KeyboardModifier.ShiftModifier.value | QtCore.Qt.KeyboardModifier.ControlModifier.value,
        "description": "Interrupt key. Responsible for interrupting the current process.",
        "signal": SignalCode.INTERRUPT_PROCESS_SIGNAL.value
    },
    {
        "display_name": "Navigate",
        "text": "Shift+Ctrl+P",
        "key": QtCore.Qt.Key.Key_P.value,
        "modifiers": QtCore.Qt.KeyboardModifier.ShiftModifier.value | QtCore.Qt.KeyboardModifier.ControlModifier.value,
        "description": "URL key. Responsible for navigating to a URL.",
        "signal": SignalCode.NAVIGATE_TO_URL.value
    },
    {
        "display_name": "Quit",
        "text": "Ctrl+Q",
        "key": QtCore.Qt.Key.Key_Q.value,
        "modifiers": QtCore.Qt.KeyboardModifier.ControlModifier.value,
        "description": "Quit key. Responsible for quitting the application.",
        "signal": SignalCode.QUIT_APPLICATION.value
    },
    {
        "display_name": "Refresh Stylesheet",
        "text": "F5",
        "key": QtCore.Qt.Key.Key_F5.value,
        "modifiers": QtCore.Qt.KeyboardModifier.NoModifier.value,
        "description": "Refresh the stylesheet. Useful when creating a template.",
        "signal": SignalCode.REFRESH_STYLESHEET_SIGNAL.value
    },
]



from typing import Callable

from datasynth.messager.publisher import PublisherMixin

from airunner.enums import SignalCode
from airunner.signal_mediator import SignalMediator
import asyncio


class MediatorMixin(PublisherMixin):
    """
    Use with any class that needs to emit and receive signals.
    Initialize with a SignalMediator instance.
    """
    def __init__(self):
        PublisherMixin.__init__(self, subject="llm_action_queue")
        self.threads = []
        self.workers = []
        self.mediator = SignalMediator()
        self.publisher_signals = []

    def emit_signal(
        self,
        code: SignalCode,
        data: object = None
    ):
        self.mediator.emit_signal(code, data)

    def register(
        self,
        code: SignalCode,
        slot_function: Callable
    ):
        """
        Accessor method for SignalMediator. register method.
        :param code:
        :param slot_function:
        :return:
        """
        self.mediator.register(code, slot_function)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'custom_llm.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QGridLayout,
    QHBoxLayout, QLineEdit, QPushButton, QScrollArea,
    QSizePolicy, QVBoxLayout, QWidget)

class Ui_custom_llm_model_widget(object):
    def setupUi(self, custom_llm_model_widget):
        if not custom_llm_model_widget.objectName():
            custom_llm_model_widget.setObjectName(u"custom_llm_model_widget")
        custom_llm_model_widget.resize(503, 300)
        self.gridLayout = QGridLayout(custom_llm_model_widget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.scan_for_models_button = QPushButton(custom_llm_model_widget)
        self.scan_for_models_button.setObjectName(u"scan_for_models_button")
        font = QFont()
        font.setPointSize(8)
        self.scan_for_models_button.setFont(font)

        self.gridLayout.addWidget(self.scan_for_models_button, 4, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.comboBox = QComboBox(custom_llm_model_widget)
        self.comboBox.setObjectName(u"comboBox")

        self.horizontalLayout.addWidget(self.comboBox)

        self.toggle_all = QCheckBox(custom_llm_model_widget)
        self.toggle_all.setObjectName(u"toggle_all")
        self.toggle_all.setFont(font)

        self.horizontalLayout.addWidget(self.toggle_all)


        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.custom_scroll_area = QScrollArea(custom_llm_model_widget)
        self.custom_scroll_area.setObjectName(u"custom_scroll_area")
        self.custom_scroll_area.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 483, 189))
        self.verticalLayout_2 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.custom_scroll_area.setWidget(self.scrollAreaWidgetContents)

        self.gridLayout.addWidget(self.custom_scroll_area, 3, 0, 1, 1)

        self.lineEdit = QLineEdit(custom_llm_model_widget)
        self.lineEdit.setObjectName(u"lineEdit")

        self.gridLayout.addWidget(self.lineEdit, 2, 0, 1, 1)


        self.retranslateUi(custom_llm_model_widget)
        self.scan_for_models_button.clicked.connect(custom_llm_model_widget.action_button_clicked_scan_for_models)
        self.comboBox.currentTextChanged.connect(custom_llm_model_widget.mode_type_changed)
        self.toggle_all.toggled.connect(custom_llm_model_widget.toggle_all_toggled)
        self.lineEdit.textEdited.connect(custom_llm_model_widget.search_text_edited)

        QMetaObject.connectSlotsByName(custom_llm_model_widget)
    # setupUi

    def retranslateUi(self, custom_llm_model_widget):
        custom_llm_model_widget.setWindowTitle(QCoreApplication.translate("custom_llm_model_widget", u"Form", None))
        self.scan_for_models_button.setText(QCoreApplication.translate("custom_llm_model_widget", u"Scan for models", None))
        self.toggle_all.setText(QCoreApplication.translate("custom_llm_model_widget", u"Toggle all", None))
        self.lineEdit.setPlaceholderText(QCoreApplication.translate("custom_llm_model_widget", u"Search models", None))
    # retranslateUi




import os
import requests
from json.decoder import JSONDecodeError
from PySide6.QtCore import QThread
from airunner.aihandler.logger import Logger
from airunner.aihandler.stablediffusion.civit_ai_download_worker import CivitAIDownloadWorker
from airunner.aihandler.stablediffusion.download_worker import DownloadWorker
from airunner.enums import SignalCode
from airunner.mediator_mixin import MediatorMixin
from airunner.windows.main.settings_mixin import SettingsMixin

logger = Logger(prefix="DownloadCivitAI")

class DownloadCivitAI(
    MediatorMixin,
    SettingsMixin
):
    def __init__(self):
        MediatorMixin.__init__(self)
        SettingsMixin.__init__(self)
        super().__init__()
        self.thread = None
        self.worker = None
        self.file_name = None

    def get_json(self, model_id: str):
        # if model_id == id/name split and get the id
        if "/" in model_id:
            model_id = model_id.split("/")[0]
        url = f"https://civitai.com/api/v1/models/{model_id}?token={self.application_settings.civit_ai_api_key}"
        headers = {
            "Content-Type": "application/json",
            #"Authorization": f"Bearer {api_token}"
        }
        logger.debug(f"Getting model data from CivitAI {url}")
        response = requests.get(url, headers=headers, allow_redirects=True)
        json = None
        try:
            json = response.json()
        except JSONDecodeError:
            print(f"Failed to decode JSON from {url}")
            print(response)
        return json

    def download_model(self, url, file_name, size_kb, callback):
        self.file_name = file_name
        self.worker = CivitAIDownloadWorker()
        self.worker.add_to_queue((
            url, file_name, size_kb
        ))
        self.thread = QThread()
        self.worker.moveToThread(self.thread)

        # Connect signals
        self.worker.finished.connect(lambda: self.emit_signal(SignalCode.DOWNLOAD_COMPLETE))
        self.worker.finished.connect(self.thread.quit)
        self.worker.progress.connect(lambda current, total: callback(current, total))
        logger.debug(f"Starting model download thread")
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.thread.started.connect(self.worker.download)
        self.thread.start()

    def remove_file(self):
        if os.path.exists(self.file_name):  # Check if the file exists and delete if so
            os.remove(self.file_name)
            print(f"Download of {self.file_name} was cancelled and the file has been deleted.")

    def stop_download(self):
        if self.worker:
            self.worker.cancel()
            self.remove_file()



from airunner.enums import SignalCode
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.canvas.templates.input_image_container_ui import Ui_input_image_container
from airunner.widgets.canvas.input_image import InputImage

class InputImageContainer(BaseWidget):
    widget_class_ = Ui_input_image_container

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register(SignalCode.MASK_GENERATOR_WORKER_RESPONSE_SIGNAL, self.on_mask_generator_worker_response_signal)
        self.register(SignalCode.MASK_UPDATED, self.on_mask_generator_worker_response_signal)
        self.input_image = None

    def on_mask_generator_worker_response_signal(self, message):
        if self.input_image:
            self.input_image.on_mask_generator_worker_response_signal()

    def showEvent(self, event):
        if self.input_image is None:
            settings_key = self.settings_key
            self.input_image = InputImage(settings_key=self.settings_key)
            self.ui.tabWidget.addTab(self.input_image, "Input Image")
            label = "Image-to-Image"
            if settings_key == "controlnet_settings":
                label = "Controlnet"
                self.input_image = InputImage(settings_key=self.settings_key, use_generated_image=True)
                self.ui.tabWidget.addTab(self.input_image, "Generated Image")
            elif settings_key == "outpaint_settings":
                label = "Inpaint / Outpaint"
            self.ui.label.setText(label)

    @property
    def settings_key(self):
        return self.property("settings_key")



from airunner.windows.setup_wizard.model_setup.stable_diffusion_setup.templates.choose_version_ui import Ui_choose_model_version
from airunner.windows.setup_wizard.setup_wizard_window import BaseWizard


class ChooseModelVersion(BaseWizard):
    class_name_ = Ui_choose_model_version




font_settings_bootstrap_data = (
    {
        "name": "chat",
        "font_family": "Arial",
        "font_size": 16,
    },
    {
        "name": "primary",
        "font_family": "Arial",
        "font_size": 12,
    }
)
